from __future__ import annotations

import hashlib
import re
from collections import UserDict
from datetime import time
from enum import IntEnum
from functools import partial
from io import BytesIO
from typing import Any, TypeVar, cast, overload

from typing_extensions import TypeAlias

from ..cos.objects.base import PdfHexString, PdfName, PdfNull, PdfObject, PdfReference
from ..cos.objects.containers import PdfArray, PdfDictionary
from ..cos.objects.stream import PdfStream
from ..cos.objects.xref import (
    CompressedXRefEntry,
    FreeXRefEntry,
    InUseXRefEntry,
    PdfXRefEntry,
    PdfXRefSection,
    PdfXRefSubsection,
)
from ..exceptions import PdfParseError
from ..security.standard_handler import StandardSecurityHandler
from .serializer import PdfSerializer
from .tokenizer import PdfTokenizer

PDF_HEADER_REGEX = re.compile(rb"PDF-(?P<major>\d+).(?P<minor>\d+)")
INDIRECT_OBJ_HEADER_REGEX = re.compile(rb"(?P<num>\d+)\s+(?P<gen>\d+)\s+obj")


def generate_file_id(filename: str, content: bytes) -> PdfHexString:
    """Generates a file identifier as described in ``§ 14.4 File identifier``."""
    id_digest = hashlib.md5(time().isoformat("auto").encode())
    id_digest.update(filename.encode())
    id_digest.update(str(len(content)).encode())

    return PdfHexString(id_digest.hexdigest().encode())


class PermsAcquired(IntEnum):
    """Permissions acquired after opening or decrypting a document."""

    NONE = 0
    """No permissions acquired, document is still encrypted."""
    USER = 1
    """User permissions within the limits specified by the security handler."""
    OWNER = 2
    """Owner permissions (all permissions)."""


class FreeObject:
    def __repr__(self) -> str:
        return "free"


MapObject: TypeAlias = "PdfObject | PdfStream | FreeObject"


class ObjectMap(UserDict[int, MapObject]):
    """A mapping of object numbers to either object references, in-use objects or free objects.

    Object references included in :attr:`.ObjectMap.unresolved` are items that have not been
    requested yet. Once an object is requested, it is removed from the unresolved set and
    added to the map as is.

    Free objects are indicated with the :class:`.FreeObject` class.
    """

    def __init__(self, pdf: PdfParser) -> None:
        super().__init__()

        self._pdf = pdf
        self.initial_reference_map: dict[int, tuple[int, int]] = {}
        """A mapping of object numbers to reference tuples for the initial entries made
        when the object map is filled."""

        self.unresolved = set()
        """A set of unresolved object numbers (objects that have not been requested 
        or cached yet)."""

    def fill(self) -> None:
        """Fills the object map with the items available in the PDF's xref table."""
        self.initial_reference_map = {
            obj: (obj, gen) for (obj, gen) in self._pdf.xref.keys()
        }
        self.unresolved.clear()

        for obj, gen in self.initial_reference_map.values():
            entry = self._pdf.xref[(obj, gen)]
            if isinstance(entry, FreeXRefEntry):
                self[obj] = FreeObject()
            else:
                self[obj] = PdfReference(obj, gen).with_resolver(self._pdf.get_object)
                self.unresolved.add(obj)

    T = TypeVar("T")

    def _make_ref(self) -> PdfReference:
        """Creates a new reference based on the current object number of the map."""
        if not self:
            return PdfReference(1, 0)

        highest_objnum = max(self.keys())
        return PdfReference(highest_objnum + 1, 0)

    def add(self, pdf_object: MapObject) -> PdfReference[MapObject]:
        """Adds a new object to the map. Returns its reference."""
        reference = self._make_ref()
        self[reference.object_number] = pdf_object

        return reference.with_resolver(self._pdf.get_object)

    def delete(self, obj_num: int) -> MapObject | None:
        """Deletes object with number ``obj_num``. Returns the object if it exists,
        otherwise returns None."""
        return self.pop(obj_num, None)

    def free(self, obj_num: int) -> None:
        """Marks object with number ``obj_num`` as a free object."""
        self[obj_num] = FreeObject()

    def __getitem__(self, number: int) -> PdfObject | PdfStream | FreeObject:
        value = super().__getitem__(number)

        if isinstance(value, PdfReference) and number in self.unresolved:
            resolved = self[number] = value.get()
            self.unresolved.discard(number)
            return resolved

        return value


class PdfParser:
    """A parser that can completely parse a PDF document.

    It consumes the PDF's cross-reference tables and trailers. It merges the tables
    into a single one and provides an interface to individually parse each indirect
    object using :class:`~pdfnaut.cos.tokenizer.PdfTokenizer`.

    Arguments:
        data (bytes):
            The document to be processed.

    Keyword Arguments:
        strict (bool, optional):
            Whether to warn or fail on issues caused by non-spec-compliance.
            Defaults to False.
    """

    def __init__(self, data: bytes, *, strict: bool = False) -> None:
        self.strict = strict
        self._tokenizer = PdfTokenizer(data)
        self._tokenizer.resolver = self.get_object

        #   object number:  direct object
        self.objects = ObjectMap(self)
        """A mapping of objects present in the document."""

        self.updates: list[PdfXRefSection] = []
        """A list of all incremental updates present in the document (most recent update first)."""

        # placeholder to make the type checker happy
        self.trailer = PdfDictionary[str, PdfObject](
            {"Size": 0, "Root": PdfReference(0, 0)}
        )
        """The most recent trailer in the PDF document.
        
        For details on the contents of the trailer, see ``§ 7.5.5 File Trailer``.
        """

        self.xref: dict[tuple[int, int], PdfXRefEntry] = {}
        """A cross-reference mapping combining the entries of all XRef tables present 
        in the document.
        
        The key is a tuple of two integers: object number and generation number. 
        The value is any of the 3 types of XRef entries (free, in use, compressed).

        This attribute reflects the state of the XRef table when the document was 
        first loaded. Assume read-only.
        """

        self.header_version = ""
        """The document's PDF version as seen in the header.

        This value should be used if no Version entry exists in the document catalog or 
        if the header's version is newer. Otherwise, use the Version entry.
        """

        self.security_handler = None
        """The document's standard security handler, if any, as specified in the Encrypt 
        dictionary of the PDF trailer.

        This field being set indicates that a supported security handler was used for
        encryption. If not set, the parser will not attempt to decrypt this document.
        """

        self._encryption_key = None

    # These are helper methods that can probably be moved out from here.

    T = TypeVar("T")

    def _ensure_object(self, obj: PdfReference[T] | T) -> T:
        return self.get_object(obj) if isinstance(obj, PdfReference) else obj

    def _get_closest(self, values: list[int], target: int) -> int:
        return min(values, key=lambda offset: abs(offset - target))

    def _tok_matches_object_header(self) -> re.Match[bytes] | None:
        if not self._tokenizer.peek().isdigit():
            return

        return re.match(INDIRECT_OBJ_HEADER_REGEX, self._tokenizer.peek_line())

    def parse(self, start_xref: int | None = None) -> None:
        """Parses the entire document.

        It begins by parsing the most recent XRef table and trailer. If this trailer
        points to a previous XRef, this function is called again with a ``start_xref``
        offset until no more XRefs are found.

        It also sets up the Standard security handler for use in case the document
        is encrypted.

        Arguments:
            start_xref (int, optional):
                The offset where the most recent XRef can be found.
        """
        # Move back for the header
        self._tokenizer.position = 0
        self.header_version = self.parse_header()

        # Because the function may be called recursively, we check if this is the first call.
        if start_xref is None:
            start_xref = self.lookup_xref_start()

        # Move to the offset where the XRef and trailer are
        self._tokenizer.position = start_xref
        section = self.parse_xref_and_trailer()

        self.updates.append(section)

        if "Prev" in section.trailer:
            # More XRefs were found. Recurse!
            self._tokenizer.position = 0
            self.parse(cast(int, section.trailer["Prev"]))
        else:
            # That's it. Merge them together.
            self.xref = self.get_merged_xrefs()
            self.trailer = self.updates[0].trailer

        # Fills the object store so we can refer to objects now!
        self.objects.fill()

        # Is the document encrypted with a standard security handler?
        if "Encrypt" in self.trailer:
            assert "ID" in self.trailer
            encryption = cast(PdfDictionary, self.trailer["Encrypt"])

            if cast(PdfName, encryption["Filter"]).value == b"Standard":
                self.security_handler = StandardSecurityHandler(
                    encryption, cast("list[PdfHexString | bytes]", self.trailer["ID"])
                )

    def parse_header(self) -> str:
        """Parses the %PDF-n.m header that is expected to be at the start of a PDF file."""
        header = self._tokenizer.parse_comment()
        pattern = re.compile(PDF_HEADER_REGEX)
        if mat := pattern.match(header.value):
            return f"{mat.group('major').decode()}.{mat.group('minor').decode()}"

        # Although not recommended, it is expected that some documents may start with
        # characters different than those of %PDF-n.m. Offsets should be calculated
        # based on the start of this token rather than the start of the document.
        if not self.strict:
            if mat := pattern.search(self._tokenizer.data):
                if self._tokenizer.data[mat.start() - 1] == 37:  # %
                    self._tokenizer.data = self._tokenizer.data[mat.start() - 1 :]
                    return (
                        f"{mat.group('major').decode()}.{mat.group('minor').decode()}"
                    )

        raise PdfParseError("Expected PDF header at start of file.")

    def build_xref_map(
        self, subsections: list[PdfXRefSubsection]
    ) -> dict[tuple[int, int], PdfXRefEntry]:
        """Creates a dictionary mapping references to XRef entries in the document."""
        entry_map: dict[tuple[int, int], PdfXRefEntry] = {}

        for subsection in subsections:
            for idx, entry in enumerate(
                subsection.entries, subsection.first_obj_number
            ):
                if isinstance(entry, FreeXRefEntry):
                    gen = entry.gen_if_used_again
                elif isinstance(entry, InUseXRefEntry):
                    gen = entry.generation
                else:
                    # compressed entries are assumed 0
                    gen = 0

                entry_map[(idx, gen)] = entry

        return entry_map

    def get_merged_xrefs(self) -> dict[tuple[int, int], PdfXRefEntry]:
        """Combines all XRef updates in the document into a cross-reference mapping
        that includes all entries."""
        entry_map: dict[tuple[int, int], PdfXRefEntry] = {}
        hybrid_objnums = []

        # from least recent to most recent
        for section in self.updates[::-1]:
            update_map = self.build_xref_map(section.subsections)

            # if the document is a hybrid-reference file, append any hidden objects.
            if "XRefStm" in section.trailer:
                self._tokenizer.position = cast(int, section.trailer["XRefStm"])

                xrefstm = self.parse_compressed_xref()
                hybrid_map = self.build_xref_map(xrefstm.subsections)

                for (obj, gen), hybrid_entry in hybrid_map.items():
                    update_entry = update_map.get((obj, gen))

                    # But only append if they aren't a thing or they are marked as "free"
                    if update_entry is None or (
                        isinstance(update_entry, FreeXRefEntry)
                        and hybrid_entry is not None
                    ):
                        entry_map[(obj, gen)] = hybrid_entry
                        hybrid_objnums.append(obj)

            entry_map.update(update_map)

        # If entries from the "hybrid section" were added, we have to remove
        # the free entries they are meant to replace. Otherwise, the object store
        # might get a bit confused and panic.
        for objnum in hybrid_objnums:
            for (num, gen), entry in entry_map.items():
                if num == objnum and isinstance(entry, FreeXRefEntry):
                    del entry_map[(num, gen)]
                    break

        return entry_map

    def lookup_xref_start(self) -> int:
        """Scans through the PDF until it finds the XRef offset then returns it"""
        contents = bytearray()

        # The PDF spec tells us we need to parse from the end of the file
        # and the XRef comes first
        self._tokenizer.position = len(self._tokenizer.data) - 1

        while self._tokenizer.position > 0:
            contents.insert(0, ord(self._tokenizer.peek()))
            if contents.startswith(b"startxref"):
                break
            self._tokenizer.position -= 1

        if not contents.startswith(b"startxref"):
            raise PdfParseError("Cannot locate XRef table. 'startxref' offset missing.")

        # advance to the startxref offset, we know it's there.
        self._tokenizer.skip(9)
        self._tokenizer.skip_whitespace()

        return int(self._tokenizer.parse_numeric())  # startxref

    def parse_xref_and_trailer(self) -> PdfXRefSection:
        """Parses both the cross-reference table and the PDF trailer.

        PDFs may include a typical uncompressed XRef table (and hence separate XRefs and
        trailers) or an XRef stream that combines both.
        """
        if self._tokenizer.matches(b"xref"):
            xref = self.parse_simple_xref()
            self._tokenizer.skip_whitespace()
            trailer = self.parse_simple_trailer()

            return PdfXRefSection(xref, trailer)
        elif self._tok_matches_object_header():
            return self.parse_compressed_xref()
        elif not self.strict:
            # let's attempt to locate a nearby xref table
            target = self._tokenizer.position
            table_offsets = self._find_xref_offsets()

            # get the xref table nearest to our offset
            self._tokenizer.position = self._get_closest(table_offsets, target)
            section = self.parse_xref_and_trailer()

            # make sure the user can see our corrections
            if "Prev" in section.trailer:
                section.trailer["Prev"] = self._get_closest(
                    table_offsets, cast(int, section.trailer["Prev"])
                )

            return section
        else:
            raise PdfParseError("XRef offset does not point to XRef section.")

    def _find_xref_offsets(self) -> list[int]:
        table_offsets = []

        # looks for the start of a xref table
        for mat in re.finditer(
            rb"(?<!start)xref(\W*)(\d+) (\d+)", self._tokenizer.data
        ):
            table_offsets.append(mat.start())

        # looks for indirect objects, then checks if they are xref streams
        for mat in re.finditer(INDIRECT_OBJ_HEADER_REGEX, self._tokenizer.data):
            self._tokenizer.position = mat.start()
            self._tokenizer.skip(mat.end() - mat.start())
            self._tokenizer.skip_whitespace()

            if self._tokenizer.matches(b"<<"):
                mapping = self._tokenizer.parse_dictionary()
                if (
                    isinstance(typ := mapping.get("Type"), PdfName)
                    and typ.value == b"XRef"
                ):
                    table_offsets.append(mat.start())

        return sorted(table_offsets)

    def parse_simple_trailer(self) -> PdfDictionary:
        """Parses the PDF's standard trailer which is used to quickly locate other
        cross reference tables and special objects.

        The trailer is separate if the XRef table is standard (uncompressed).
        Otherwise it is part of the XRef object."""
        self._tokenizer.skip(7)  # past the 'trailer' keyword
        self._tokenizer.skip_whitespace()

        # next token is a dictionary
        return self._tokenizer.parse_dictionary()

    def parse_simple_xref(self) -> list[PdfXRefSubsection]:
        """Parses a standard, uncompressed XRef table of the format described in
        ``§ 7.5.4 Cross-Reference Table``.

        If ``startxref`` points to an XRef object, :meth:`.parse_compressed_xref`
        should be called instead.
        """
        self._tokenizer.skip(4)
        self._tokenizer.skip_whitespace()

        subsections = []

        while not self._tokenizer.done:
            # subsection
            subsection = re.match(
                rb"(?P<first_obj>\d+)\s(?P<count>\d+)", self._tokenizer.peek_line()
            )
            if subsection is None:
                break

            self._tokenizer.skip(subsection.end())
            self._tokenizer.skip_whitespace()

            # xref entries
            entries: list[PdfXRefEntry] = []
            for idx in range(int(subsection.group("count"))):
                entry = re.match(
                    rb"(?P<offset>\d{10}) (?P<gen>\d{5}) (?P<status>f|n)",
                    self._tokenizer.peek(20),
                )
                if entry is None:
                    raise PdfParseError(f"Expected valid XRef entry at row {idx + 1}")

                offset = int(entry.group("offset"))
                generation = int(entry.group("gen"))

                if entry.group("status") == b"n":
                    entries.append(InUseXRefEntry(offset, generation))
                else:
                    entries.append(FreeXRefEntry(offset, generation))

                # some files do not respect the 20-byte length req. for entries
                # hence this is here for tolerance
                self._tokenizer.skip(entry.end())
                self._tokenizer.skip_whitespace()

            subsections.append(
                PdfXRefSubsection(
                    int(subsection.group("first_obj")),
                    int(subsection.group("count")),
                    entries,
                )
            )

        return subsections

    def parse_compressed_xref(self) -> PdfXRefSection:
        """Parses a compressed cross-reference stream which includes both the XRef table
        and information from the PDF trailer.

        Described in ``§ 7.5.8 Cross-Reference Streams``."""
        xref_stream = self.parse_indirect_object(
            InUseXRefEntry(self._tokenizer.position, 0), None
        )
        assert isinstance(xref_stream, PdfStream)

        contents = BytesIO(xref_stream.decode())

        xref_widths = cast(PdfArray[int], xref_stream.details["W"])
        xref_indices = cast(
            PdfArray[int],
            xref_stream.details.get(
                "Index", PdfArray([0, xref_stream.details["Size"]])
            ),
        )

        subsections = []

        for idx in range(0, len(xref_indices), 2):
            subsection = PdfXRefSubsection(
                first_obj_number=xref_indices[idx],
                count=xref_indices[idx + 1],
                entries=[],
            )

            for _ in range(subsection.count):
                field_type = int.from_bytes(
                    contents.read(xref_widths[0]) or b"\x01", "big"
                )
                second = int.from_bytes(contents.read(xref_widths[1]), "big")
                third = int.from_bytes(contents.read(xref_widths[2]), "big")

                # TODO: Warn of unknown field types once we have a logging facility
                if field_type == 0:
                    subsection.entries.append(
                        FreeXRefEntry(next_free_object=second, gen_if_used_again=third)
                    )
                elif field_type == 1:
                    subsection.entries.append(
                        InUseXRefEntry(offset=second, generation=third)
                    )
                elif field_type == 2:
                    subsection.entries.append(
                        CompressedXRefEntry(objstm_number=second, index_within=third)
                    )

            subsections.append(subsection)

        return PdfXRefSection(subsections, xref_stream.details)

    def parse_indirect_object(
        self, xref_entry: InUseXRefEntry, reference: PdfReference | None
    ) -> PdfObject | PdfStream:
        """Parses an indirect object not within an object stream, or basically, an object
        that is directly referred to by an ``xref_entry`` and a ``reference``"""
        self._tokenizer.position = xref_entry.offset
        self._tokenizer.skip_whitespace()

        mat = self._tok_matches_object_header()
        if not mat:
            raise PdfParseError("XRef entry does not point to indirect object.")
        self._tokenizer.skip(mat.end())
        self._tokenizer.skip_whitespace()

        contents = self._tokenizer.get_next_token()
        self._tokenizer.skip_whitespace()

        # uh oh, a stream?
        if self._tokenizer.matches(b"stream"):
            extent = cast(PdfDictionary, contents)
            # the implicit get_object call might move us around so we must save and then
            # restore the previous position
            _current = self._tokenizer.position
            length = extent["Length"]
            self._tokenizer.position = _current

            if not isinstance(length, int):
                raise PdfParseError("\\Length entry of stream extent not an integer")

            item = PdfStream(extent, self.parse_stream(xref_entry, length))
        else:
            item = cast(PdfObject, contents)

        return self._get_decrypted(item, reference)

    @overload
    def _get_decrypted(
        self, pdf_object: PdfObject, reference: PdfReference | None
    ) -> PdfObject: ...

    @overload
    def _get_decrypted(
        self, pdf_object: PdfStream, reference: PdfReference | None
    ) -> PdfStream: ...

    def _get_decrypted(
        self, pdf_object: PdfObject | PdfStream, reference: PdfReference | None
    ) -> PdfObject | PdfStream:
        if (
            self.security_handler is None
            or not self._encryption_key
            or reference is None
        ):
            return pdf_object

        if isinstance(pdf_object, PdfStream):
            use_stmf = True

            # Don't use StmF if the stream handles its own encryption
            if filter_ := pdf_object.details.get("Filter"):
                if isinstance(filter_, PdfName):
                    filters = PdfArray[PdfName]([filter_])
                else:
                    filters = cast(PdfArray[PdfName], filter_)

                for name in filters:
                    if name.value == b"Crypt":
                        use_stmf = False
                        pdf_object._crypt_params = {
                            "Handler": self.security_handler,
                            "EncryptionKey": self._encryption_key,
                            "Reference": reference,
                        }
                        break

            if use_stmf:
                pdf_object.raw = self.security_handler.decrypt_object(
                    self._encryption_key, pdf_object, reference
                )

            return pdf_object
        elif isinstance(pdf_object, PdfHexString):
            return PdfHexString.from_raw(
                self.security_handler.decrypt_object(
                    self._encryption_key, pdf_object.value, reference
                )
            )
        elif isinstance(pdf_object, bytes):
            return self.security_handler.decrypt_object(
                self._encryption_key, pdf_object, reference
            )
        elif isinstance(pdf_object, PdfArray):
            return PdfArray(
                self._get_decrypted(obj, reference) for obj in pdf_object.data
            )
        elif isinstance(pdf_object, PdfDictionary):
            # The Encrypt key does not need decrypting.
            if reference == self.trailer.data["Encrypt"]:
                return pdf_object

            return PdfDictionary(
                {
                    name: self._get_decrypted(cast(PdfObject, value), reference)
                    for name, value in pdf_object.data.items()
                }
            )

        # Why would a number be encrypted?
        return pdf_object

    def parse_stream(self, xref_entry: InUseXRefEntry, extent: int) -> bytes:
        """Parses a PDF stream of length ``extent``"""
        self._tokenizer.skip(6)  # past the 'stream' keyword
        self._tokenizer.skip_next_eol(no_cr=True)

        contents = self._tokenizer.consume(extent)
        self._tokenizer.skip_next_eol(no_cr=True)

        if self.xref:
            # We get the offset of the entry directly following this one as a bounds check
            next_entry_at = iter(
                val
                for val in self.xref.values()
                if isinstance(val, InUseXRefEntry) and val.offset > xref_entry.offset
            )
        else:
            # The stream being parsed is (most likely) part of an XRef object
            next_entry_at = iter([])

        # Have we gone way beyond the stream?
        try:
            if self._tokenizer.position >= next(next_entry_at).offset:
                raise PdfParseError(
                    "\\Length key in stream extent parses beyond object."
                )
        except StopIteration:
            pass

        self._tokenizer.skip_whitespace()
        # Are we done?
        if not self._tokenizer.skip_if_matches(b"endstream"):
            raise PdfParseError(
                "\\Length key in stream extent does not match end of stream."
            )

        return contents

    T = TypeVar("T")

    @overload
    def get_object(self, reference: PdfReference[T], cache: bool = True) -> T: ...

    @overload
    def get_object(
        self, reference: tuple[int, int], cache: bool = True
    ) -> PdfObject | PdfStream | PdfNull | FreeObject: ...

    def get_object(
        self, reference: PdfReference | tuple[int, int], cache: bool = True
    ) -> PdfObject | PdfStream | PdfNull | FreeObject | Any:
        """Resolves a reference into the indirect object it points to.

        Arguments:
            reference (:class:`.PdfReference` | :class:`tuple[int, int]`):
                A :class:`.PdfReference` object or a tuple of two integers representing,
                in order, the object number and the generation number.

            cache (bool, optional):
                Whether to interact with the object store when resolving references.
                Defaults to True.

                When True, the parser will read entries from the object store and write new
                ones if they are not present. If False, the parser will always fetch new
                entries and will not write to the object store.

                Note that the object store will be accessed regardless of the value of
                ``cache``if the object is new and is not included in the xref table.

        Returns:
            The object the reference resolves to.

            If the reference is invalid (i.e. does not exist), returns :class:`.PdfNull`.
            If the object referred to is a free object, returns :class:`.FreeObject`.
        """
        if isinstance(reference, tuple):
            reference = PdfReference(*reference).with_resolver(self.get_object)

        # If cache requested and the object hasn't been resolved or cached yet.
        if cache and reference.object_number not in self.objects.unresolved:
            return self.objects.get(reference.object_number)

        root_entry = self.xref.get((reference.object_number, reference.generation))
        if root_entry is None:
            if (obj_entry := self.objects.get(reference.object_number)) is not None:
                return obj_entry

            return PdfNull()

        if isinstance(root_entry, InUseXRefEntry):
            obj = self.parse_indirect_object(root_entry, reference)
            if not cache:
                return obj

            # Add to cache then set the object as resolved.
            self.objects[reference.object_number] = obj
            self.objects.unresolved.discard(reference.object_number)

            return self.objects[reference.object_number]
        elif isinstance(root_entry, CompressedXRefEntry):
            # TODO: Add support for editing object streams
            # Get the object stream it's part of (gen always 0)
            objstm_ref = (root_entry.objstm_number, 0)
            objstm_entry = self.xref[objstm_ref]
            assert isinstance(objstm_entry, InUseXRefEntry)

            if cache and root_entry.objstm_number not in self.objects.unresolved:
                objstm = self.objects[root_entry.objstm_number]
            else:
                objstm = self.parse_indirect_object(
                    objstm_entry,
                    PdfReference(*objstm_ref).with_resolver(
                        partial(self.get_object, cache=False)
                    ),
                )

            assert isinstance(objstm, PdfStream)

            if cache:
                self.objects[root_entry.objstm_number] = objstm
                self.objects.unresolved.discard(root_entry.objstm_number)

            seq = PdfTokenizer(objstm.decode()[objstm.details["First"] :] or b"")
            seq.resolver = self.get_object

            for idx, token in enumerate(seq):
                if idx == root_entry.index_within:
                    return token

        return PdfNull()

    def decrypt(self, password: str) -> PermsAcquired:
        """Decrypts this document through the Standard security handler using the
        provided ``password``.

        The standard security handler may specify 2 passwords: an owner password and a user
        password. The owner password would allow full access to the PDF and the user password
        should allow access according to the permissions specified in the document.

        Returns:
            A :class:`.PermsAcquired` specifying the permissions acquired by ``password``.

            - If the document is not encrypted, defaults to :attr:`.PermsAcquired.OWNER`
            - if the document was not decrypted, defaults to :attr:`.PermsAcquired.NONE`
        """
        if self.security_handler is None:
            return PermsAcquired.OWNER

        # Is this the owner password?
        encryption_key, is_owner_pass = (
            self.security_handler.authenticate_owner_password(password.encode())
        )
        if is_owner_pass:
            self._encryption_key = encryption_key
            return PermsAcquired.OWNER

        # Is this the user password
        encryption_key, is_user_pass = self.security_handler.authenticate_user_password(
            password.encode()
        )
        if is_user_pass:
            self._encryption_key = encryption_key
            return PermsAcquired.USER

        return PermsAcquired.NONE

    def save(self, filename: str) -> None:
        """Saves the contents of this parser to ``filename``."""
        builder = PdfSerializer()
        builder.write_header("2.0")

        rows: list[tuple[int, PdfXRefEntry]] = []

        use_compressed = False
        update_freelist = False

        for obj_num in self.objects:
            ref_tup = self.objects.initial_reference_map.get(obj_num)
            # Object is new
            if ref_tup is None:
                resolved = self.objects[obj_num]
                if isinstance(resolved, FreeObject):
                    rows.append((obj_num, FreeXRefEntry(-1, 0)))
                    update_freelist = True
                else:
                    offset = builder.write_object((obj_num, 0), resolved)
                    rows.append((obj_num, InUseXRefEntry(offset, 0)))

                continue

            # Object is modified or left intact
            entry = self.xref[ref_tup]
            if isinstance(entry, FreeXRefEntry):
                resolved = self.objects[obj_num]
                # Free entry left unmodified or can no longer be used
                if isinstance(resolved, FreeObject) or entry.gen_if_used_again >= 65535:
                    rows.append(
                        (
                            obj_num,
                            FreeXRefEntry(
                                entry.next_free_object, entry.gen_if_used_again
                            ),
                        )
                    )
                    continue

                # Free entry now in use
                offset = builder.write_object(
                    (obj_num, entry.gen_if_used_again), resolved
                )
                rows.append((obj_num, InUseXRefEntry(offset, entry.gen_if_used_again)))
                update_freelist = True
            elif isinstance(entry, InUseXRefEntry):
                resolved = self.objects[obj_num]
                # In use object freed
                if isinstance(resolved, FreeObject):
                    rows.append((obj_num, FreeXRefEntry(-1, entry.generation + 1)))
                    update_freelist = True
                    continue

                # In use object either modified or left intact
                offset = builder.write_object(ref_tup, resolved)
                rows.append((obj_num, InUseXRefEntry(offset, entry.generation)))
            elif isinstance(entry, CompressedXRefEntry):
                # TODO: Add support for compressed entries (aka object streams)
                use_compressed = True
                rows.append(
                    (
                        obj_num,
                        CompressedXRefEntry(entry.objstm_number, entry.index_within),
                    )
                )

        if update_freelist:
            # let's first get the members of the freelist
            freelist_members = [
                idx
                for idx, entry in enumerate(rows)
                if isinstance(entry, FreeXRefEntry)
            ]

            for freelist_idx, xref_idx in enumerate(freelist_members):
                obj_num, entry = rows[xref_idx]
                assert isinstance(entry, FreeXRefEntry)

                if freelist_idx + 1 < len(freelist_members):
                    entry.next_free_object = rows[freelist_members[freelist_idx + 1]][0]
                else:
                    entry.next_free_object = 0

                rows[xref_idx] = (obj_num, entry)

        xref_section = builder.generate_xref_section(rows)

        new_trailer = PdfDictionary(
            {
                "Size": len(self.build_xref_map(xref_section)),
                "Root": self.trailer.data["Root"],
            }
        )

        if "Info" in self.trailer.data:
            new_trailer.data["Info"] = self.trailer.data["Info"]

        if "ID" in self.trailer.data:
            ids = cast(PdfArray[PdfHexString | bytes], self.trailer.data["ID"])
            new_trailer.data["ID"] = PdfArray(
                [ids[0], generate_file_id(filename, builder.content)]
            )

        if use_compressed:
            startxref = builder.write_compressed_xref_section(
                PdfXRefSection(xref_section, new_trailer)
            )
            builder.write_trailer(None, startxref)
        else:
            startxref = builder.write_standard_xref_section(xref_section)
            builder.write_trailer(new_trailer, startxref)

        builder.write_eof()

        with open(filename, "wb") as output_fp:
            output_fp.write(builder.content)



from __future__ import annotations

import datetime
import enum
from typing import Any, Generic, Type, TypeVar, cast

from pdfnaut.cos.objects.base import (
    PdfHexString,
    PdfName,
    PdfObject,
    encode_text_string,
    parse_text_string,
)
from pdfnaut.cos.objects.containers import PdfDictionary
from pdfnaut.cos.objects.date import PdfDate

K = TypeVar("K")
V = TypeVar("V")


class Required:
    """Sentinel to mark a field as required."""

    pass


class TextStringField:
    """A field defining a key whose value is a text string (``§ 7.9.2.2 Text string type``)."""

    def __init__(self, field: str) -> None:
        self.field = field

    def __get__(self, obj: PdfDictionary, objtype: Any | None = None) -> str | None:
        if (value := obj.get(self.field)) is not None:
            return parse_text_string(cast("PdfHexString | bytes", value))

    def __set__(self, obj: PdfDictionary, value: str) -> None:
        obj[self.field] = encode_text_string(value)

    def __delete__(self, obj: PdfDictionary) -> None:
        obj.pop(self.field, None)


ET = TypeVar("ET")
LT = TypeVar("LT")


class EnumField(Generic[LT, ET]):
    """A field defining a key whose value is a set of names part of an enumeration."""

    def __init__(self, field: str, enum_map: dict[LT, ET], default: ET) -> None:
        self.field = field
        self.default = default
        self.enum_map = enum_map

    def __get__(self, obj: PdfDictionary, objtype: Any | None = None) -> ET:
        if (value := obj.get(self.field)) is not None:
            name = cast(LT, cast(PdfName, value).value.decode())
            return self.enum_map[name]

        return self.default

    def __set__(self, obj: PdfDictionary, value: ET) -> None:
        lit_map = {val: key for key, val in self.enum_map.items()}
        obj[self.field] = PdfName(lit_map[value].encode())  # type: ignore

    def __delete__(self, obj: PdfDictionary) -> None:
        obj.pop(self.field, None)


T = TypeVar("T")


class StandardField(Generic[T]):
    """A field defining a key whose value is one of the following basic types: boolean,
    numeric, arrays, dictionaries, names, streams, and null.

    Text strings and dates have special handling and are better served by the
    :class:`.TextStringField` and :class:`.DateField` classes respectively.

    Names part of an enumeration are better served by the :class:`.EnumField` class."""

    def __init__(self, field: str, default: T | Type[Required] = Required) -> None:
        self.field = field
        self.default = default

    def __get__(self, obj: PdfDictionary, objtype: Any | None = None) -> T:
        if self.default is Required:
            return cast(T, obj[self.field])

        return cast(T, obj.get(self.field, self.default))

    def __set__(self, obj: PdfDictionary, value: T) -> None:
        obj[self.field] = cast(PdfObject, value)

    def __delete__(self, obj: PdfDictionary) -> None:
        obj.pop(self.field, None)


T = TypeVar("T")


class NameField(Generic[T]):
    """A field defining a key whose value is a name."""

    def __init__(self, field: str, default: T | Type[Required] = Required) -> None:
        self.field = field
        self.default = default

    def __get__(self, obj: PdfDictionary, objtype: Any | None = None) -> T:
        if self.default is Required:
            name = cast(PdfName, obj[self.field]).value.decode()
            return cast(T, name)

        name = cast(PdfName, obj.get(self.field, self.default)).value.decode()
        return cast(T, name)

    def __set__(self, obj: PdfDictionary, value: T) -> None:
        obj[self.field] = PdfName(cast(str, value).encode())

    def __delete__(self, obj: PdfDictionary) -> None:
        obj.pop(self.field, None)


E = TypeVar("E", bound=enum.IntFlag)


class FlagField(Generic[E]):
    """A field defining a key whose value is part of a set of bit flags."""

    def __init__(
        self, field: str, enum_cls: Type[E], default: E | Type[Required] = Required
    ) -> None:
        self.field = field
        self.enum_cls = enum_cls
        self.default = default

    def __get__(self, obj: PdfDictionary, objtype: Any | None = None) -> E:
        if self.default is Required:
            return self.enum_cls(obj[self.field])

        value = obj.get(self.field)
        if value is None:
            return cast(E, self.default)

        return self.enum_cls(value)

    def __set__(self, obj: PdfDictionary, value: E) -> None:
        obj[self.field] = int(value)

    def __delete__(self, obj: PdfDictionary) -> None:
        obj.pop(self.field, None)


class DateField:
    """A field defining a key whose value is a date (``§ 7.9.4 Dates``)."""

    def __init__(self, field: str) -> None:
        self.field = field

    def __get__(self, obj: PdfDictionary, objtype: Any | None = None) -> PdfDate | None:
        text = TextStringField(self.field).__get__(obj)
        if text is not None:
            return PdfDate.from_pdf(text)

    def __set__(self, obj: PdfDictionary, value: PdfDate | datetime.datetime) -> None:
        if isinstance(value, datetime.datetime):
            value = PdfDate.from_datetime(value)

        TextStringField(self.field).__set__(obj, value.as_pdf_string())

    def __delete__(self, obj: PdfDictionary) -> None:
        obj.pop(self.field, None)



from __future__ import annotations

# NOTE: There's presumably some issues with how some Python 3.9 releases handle
#       inheriting from a Protocol subclass. The fix was merged in 3.10.
#       This is why this backport is used.
#       https://github.com/python/cpython/issues/89284
from typing_extensions import Protocol


class CryptProvider(Protocol):
    key: bytes

    def __init__(self, key: bytes) -> None:
        self.key = key

    def encrypt(self, contents: bytes) -> bytes: ...
    def decrypt(self, contents: bytes) -> bytes: ...


class IdentityProvider(CryptProvider):
    """The Identity provider does nothing - same input, same output"""

    def encrypt(self, contents: bytes) -> bytes:
        return contents

    def decrypt(self, contents: bytes) -> bytes:
        return contents



from __future__ import annotations

import re
from collections.abc import Callable
from typing import cast

from ..cos.objects import (
    ObjectGetter,
    PdfArray,
    PdfComment,
    PdfDictionary,
    PdfHexString,
    PdfName,
    PdfNull,
    PdfObject,
    PdfOperator,
    PdfReference,
)

# as defined in § 7.2.3 Character Set, Table 1 & Table 2
DELIMITERS = b"()<>[]{}/%"
WHITESPACE = b"\x00\t\n\x0c\r "
EOL_CR = b"\r"
EOL_LF = b"\n"
EOL_CRLF = b"\r\n"

# as defined in § 7.3.4.2 Literal Strings, Table 3
STRING_ESCAPE = {
    b"\\n": b"\n",
    b"\\r": b"\r",
    b"\\t": b"\t",
    b"\\b": b"\b",
    b"\\f": b"\f",
    b"\\(": b"(",
    b"\\)": b")",
    b"\\\\": b"\\",
}


class ContentStreamIterator:
    """An iterator designed to consume the operators of a content stream.

    For each instruction in the stream, this iterator will yield a tuple including, in order,
    the name of the operator and a list of operands.
    """

    def __init__(self, contents: bytes) -> None:
        self.contents = contents
        self.tokenizer = PdfTokenizer(contents, parse_operators=True)

    def __iter__(self):
        return self

    def __next__(self) -> tuple[str, PdfArray]:
        operands = PdfArray()

        for tok in self.tokenizer:
            if not isinstance(tok, PdfOperator):
                operands.append(cast(PdfObject, tok))
            else:
                return (tok.value.decode(), operands)

        raise StopIteration


class PdfTokenizer:
    """A tokenizer designed to consume individual objects that do not depend on a cross
    reference table. It is used by :class:`~pdfnaut.cos.parser.PdfParser` for this purpose.

    This tokenizer consumes basic objects such as arrays and dictionaries. Indirect objects
    and streams depend on an XRef table and hence are not sequentially parsable. It is not
    intended to parse these items but rather the objects stored within them.

    Arguments:
        data (bytes):
            The contents to be parsed.

        parse_operators (bool, optional):
            Whether to also parse the operators present in content streams.
            Defaults to False.
    """

    def __init__(self, data: bytes, *, parse_operators: bool = False) -> None:
        self.data = data
        self.position = 0
        self.resolver: ObjectGetter | None = None

        self.parse_operators = parse_operators

    def __iter__(self):
        return self

    def __next__(self) -> PdfObject | PdfComment | PdfOperator:
        while not self.done:
            if (tok := self.get_next_token()) is not None:
                return tok
            self.skip()
        raise StopIteration

    # * Scanning
    @property
    def done(self) -> bool:
        """Whether the parser has reached the end of data."""
        return self.position >= len(self.data)

    def skip(self, n: int = 1) -> None:
        """Skips/advances ``n`` characters in the tokenizer."""
        if not self.done:
            self.position += n

    def peek(self, n: int = 1) -> bytes:
        """Peeks ``n`` characters into ``data`` without advancing through the tokenizer."""
        return self.data[self.position : self.position + n]

    def peek_line(self) -> bytes:
        """Peeks from the current position until an EOL marker is found (not included
        in the output)."""
        start_pos = self.position
        line = self.consume_while(lambda _: not self.peek(2).startswith((EOL_CRLF, EOL_CR, EOL_LF)))
        self.position = start_pos
        return line

    def consume(self, n: int = 1) -> bytes:
        """Consumes and returns ``n`` characters."""
        consumed = self.peek(n)
        self.skip(len(consumed))

        return consumed

    def matches(self, keyword: bytes) -> bool:
        """Checks whether ``keyword`` starts at the current position."""
        return self.peek(len(keyword)) == keyword

    def _get_reference_if_matched(self) -> re.Match[bytes] | None:
        """Returns the Regex match if an indirect reference is at the current position
        otherwise None."""
        if not self.peek().isdigit():
            return

        return re.match(rb"^(?P<num>\d+)\s+(?P<gen>\d+)\s+R", self.peek_line())

    def skip_if_matches(self, keyword: bytes) -> bool:
        """Advances ``len(keyword)`` characters if ``keyword`` starts at the current
        position. Returns whether the match was successful."""
        if self.matches(keyword):
            self.skip(len(keyword))
            return True
        return False

    def skip_whitespace(self) -> None:
        """Advances through PDF whitespace."""
        self.skip_while(lambda ch: ch in WHITESPACE)

    def skip_next_eol(self, no_cr: bool = False) -> None:
        """Skips the next EOL marker if matched. If ``no_cr`` is True, CR (``\\r``) will
        not be treated as a newline."""
        matched = self.skip_if_matches(EOL_CRLF)
        if no_cr and self.matches(EOL_CR):
            return

        if not matched and self.peek() in EOL_CRLF:
            self.skip()

    def skip_while(self, callback: Callable[[bytes], bool], *, limit: int = -1) -> int:
        """Skips while ``callback`` returns True for an input character. If specified,
        it will only skip ``limit`` characters. Returns how many characters were skipped."""
        if limit == -1:
            limit = len(self.data)

        start = self.position
        while not self.done and callback(self.peek()) and self.position - start < limit:
            self.position += 1
        return self.position - start

    def consume_while(self, callback: Callable[[bytes], bool], *, limit: int = -1) -> bytes:
        """Consumes while ``callback`` returns True for an input character. If specified,
        it will only consume up to ``limit`` characters."""
        if limit == -1:
            limit = len(self.data)

        consumed = b""
        while not self.done and callback(self.peek()) and len(consumed) < limit:
            consumed += self.consume()
        return consumed

    def get_next_token(self) -> PdfObject | PdfComment | PdfOperator | None:
        """Parses and returns the next token."""
        if self.done:
            return

        if self.skip_if_matches(b"true"):
            return True
        elif self.skip_if_matches(b"false"):
            return False
        elif self.skip_if_matches(b"null"):
            return PdfNull()
        elif mat := self._get_reference_if_matched():
            return self.parse_indirect_reference(mat)
        elif self.peek().isdigit() or self.peek() in b"+-":
            return self.parse_numeric()
        elif self.matches(b"["):
            return self.parse_array()
        elif self.matches(b"/"):
            return self.parse_name()
        elif self.matches(b"<<"):
            return self.parse_dictionary()
        elif self.matches(b"<"):
            return self.parse_hex_string()
        elif self.matches(b"("):
            return self.parse_literal_string()
        elif self.matches(b"%"):
            return self.parse_comment()
        elif self.parse_operators and self.peek().isalpha() or self.peek() in b"'\"":
            return self.parse_operator()

    def parse_numeric(self) -> int | float:
        """Parses a numeric object.

        PDF has two types of numbers: integers (40, -30) and real numbers (3.14). The range
        and precision of these numbers may depend on the machine used to process the PDF.
        """
        prefix_or_digit = self.consume()  # either a digit or a sign prefix
        number = prefix_or_digit + self.consume_while(lambda ch: ch.isdigit() or ch == b".")

        # is this a float (a real number)?
        if b"." in number:
            return float(number)
        return int(number)

    def parse_name(self) -> PdfName:
        """Parses a name -- a uniquely defined atomic symbol introduced with a slash
        and ending before a delimiter or whitespace."""
        self.skip()  # past the /

        atom = b""
        while not self.done and self.peek() not in DELIMITERS + WHITESPACE:
            if self.matches(b"#"):
                # escape sequence matched
                self.skip()

                atom += chr(int(self.consume(2), 16)).encode()
                continue

            atom += self.consume()

        return PdfName(atom)

    def parse_hex_string(self) -> PdfHexString:
        """Parses a hexadecimal string. Hexadecimal strings usually include arbitrary binary
        data. If the sequence is uneven, the last character is assumed to be 0."""
        self.skip()  # adv. past the <

        content = self.consume_while(lambda ch: ch != b">")
        self.skip()  # adv. past the >

        return PdfHexString(content)

    def parse_dictionary(self) -> PdfDictionary:
        """Parses a dictionary object. In a PDF, dictionary keys are name objects and
        dictionary values are any object or reference. This parser maps name objects to
        strings in this context."""
        self.skip(2)  # adv. past the <<

        kv_pairs: list[PdfObject] = []

        while not self.done and not self.matches(b">>"):
            if (token := self.get_next_token()) is not None:
                kv_pairs.append(cast(PdfObject, token))

            # Only advance when no token matches. The individual object
            # parsers already advance and this avoids advancing past delimiters.
            if token is None:
                self.skip()

        self.skip(2)  # adv. past the >>

        return PdfDictionary(
            {
                cast(PdfName, kv_pairs[i]).value.decode(): kv_pairs[i + 1]
                for i in range(0, len(kv_pairs), 2)
            }
        )

    def parse_array(self) -> PdfArray:
        """Parses an array. Arrays are heterogenous in PDF so they are mapped to Python lists."""
        self.skip()  # past the [

        items = PdfArray[PdfObject]()

        while not self.done and not self.matches(b"]"):
            if (token := self.get_next_token()) is not None:
                items.append(cast(PdfObject, token))

            if token is None:
                self.skip()

        self.skip()  # past the ]

        return items

    def parse_indirect_reference(self, mat: re.Match[bytes]) -> PdfReference:
        """Parses an indirect reference. Indirect references allow locating an object in a PDF."""
        self.skip(mat.end())  # consume the reference
        self.skip_whitespace()

        reference = PdfReference(int(mat.group("num")), int(mat.group("gen")))
        if self.resolver:
            reference._resolver = self.resolver

        return reference

    def parse_literal_string(self) -> bytes:
        """Parses a literal string. Literal strings may be entirely ASCII or may include
        arbitrary binary data (usually when they are encrypted). This parser does not currently
        distinguish between different encodings."""
        self.skip()  # past the (

        string = b""
        # balanced parentheses do not require escaping
        paren_depth = 1

        while not self.done and paren_depth >= 1:
            if self.matches(b"\\"):
                # Is this a default escape? (Table 3 § 7.3.4.2)
                escape = STRING_ESCAPE.get(self.peek(2))
                if escape is not None:
                    string += escape
                    self.skip(2)  # past the escape code
                    continue

                # Otherwise, match a newline or a \ddd sequence
                self.skip(1)

                matched = self.skip_if_matches(EOL_CRLF)
                if not matched and self.peek() in EOL_CRLF:
                    self.skip()
                elif self.peek().isdigit():
                    octal_code = self.consume_while(bytes.isdigit, limit=3)
                    string += chr(int(octal_code, 8)).encode()

            if self.matches(b"("):
                paren_depth += 1
            elif self.matches(b")"):
                paren_depth -= 1

            # This avoids appending the delimiting paren
            if paren_depth != 0:
                string += self.peek()
            self.skip()

        return string

    def parse_comment(self) -> PdfComment:
        """Parses a PDF comment. Comments have no syntactical meaning."""
        self.skip()  # past the %

        line = self.consume_while(lambda ch: ch not in EOL_CRLF)
        self.skip_whitespace()

        return PdfComment(line)

    def parse_operator(self) -> PdfOperator:
        """Parses a PDF operator. Operators can be found in content streams and are only
        parsed if :attr:`.parse_operators` is true."""

        operator = self.consume_while(lambda ch: ch not in DELIMITERS + WHITESPACE)
        return PdfOperator(operator)



from __future__ import annotations


class PdfParseError(Exception):
    """The parser was unable to continue parsing the PDF"""

    pass


class PdfFilterError(PdfParseError):
    """A filter was unable to decode a stream or is not supported"""

    pass


class PdfResolutionError(Exception):
    """The document was unable to resolve a reference because no resolution method
    is available."""

    pass


class PdfWriteError(Exception):
    """The writer was unable to serialize an object"""

    pass



from __future__ import annotations

from typing import cast

from pdfnaut import PdfParser
from pdfnaut.cos.objects import PdfStream
from pdfnaut.filters import ASCII85Filter, ASCIIHexFilter, FlateFilter, RunLengthFilter


def test_ascii() -> None:
    assert ASCIIHexFilter().decode(b"50444673>") == b"PDFs"
    assert ASCII85Filter().decode(b":ddco~>") == b"PDFs"

    assert ASCIIHexFilter().encode(b"band") == b"62616E64>"
    assert ASCII85Filter().encode(b"band") == b"@UX.b~>"


def test_flate() -> None:
    # No predictor
    encoded_str = b"x\x9c\x0bpq+\x06\x00\x03\x0f\x01N"
    assert FlateFilter().decode(encoded_str) == b"PDFs"
    assert FlateFilter().encode(b"PDFs") == encoded_str


def test_rle() -> None:
    with open("tests/docs/river-rle-image.pdf", "rb") as fp:
        pdf = PdfParser(fp.read())
        pdf.parse()

        rle_stream = cast(PdfStream, pdf.get_object((3, 0)))

        with (
            open("tests/docs/filters/rle-input.jpg", "rb") as input_image,
            open("tests/docs/filters/rle-output.bin", "rb") as output,
        ):
            assert RunLengthFilter().decode(rle_stream.raw) == input_image.read()

            input_image.seek(0)
            assert RunLengthFilter().encode(input_image.read()) == output.read()



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, cast

from typing_extensions import Self

from ...exceptions import PdfFilterError
from ...filters import SUPPORTED_FILTERS
from .base import PdfName, PdfNull, PdfObject
from .containers import PdfArray, PdfDictionary


@dataclass
class PdfStream:
    """A sequence of bytes that may be of unlimited length. Objects with a large
    amount of data like images or fonts are usually represented by streams
    (``§ 7.3.8 Stream objects``)."""

    details: PdfDictionary[str, PdfObject]
    """The stream extent dictionary (``§ 7.3.8.2 Stream extent``)."""

    raw: bytes = field(repr=False)
    """The raw data which is assumed to be encoded."""

    _crypt_params: dict[str, Any] = field(default_factory=dict, repr=False)
    """Parameters specific to the Crypt filter"""

    def decode(self) -> bytes:
        """Returns the decoded contents of the stream. If no filter is defined,
        it returns the original contents.

        Raises :class:`.pdfnaut.exceptions.PdfFilterError` if a filter is unsupported."""

        filters = cast("PdfName | PdfArray[PdfName] | None", self.details.get("Filter"))
        params = cast("PdfDictionary | PdfArray[PdfDictionary]", self.details.get("DecodeParms"))

        if filters is None:
            return self.raw

        if isinstance(filters, PdfName):
            filters = PdfArray([filters])

        if not isinstance(params, PdfArray):
            params = PdfArray([params])

        output = self.raw

        for filt, params in zip(filters, params):
            if filt.value not in SUPPORTED_FILTERS:
                raise PdfFilterError(f"{filt.value.decode()}: Filter is unsupported.")

            if isinstance(params, PdfNull) or params is None:
                params = PdfDictionary[str, PdfObject]()

            if filt.value == b"Crypt" and self._crypt_params.get("Handler"):
                params.update(self._crypt_params)

            output = SUPPORTED_FILTERS[filt.value]().decode(self.raw, params=params)

        return output

    @classmethod
    def create(
        cls,
        raw: bytes,
        details: PdfDictionary | None = None,
        crypt_params: PdfDictionary | None = None,
    ) -> Self:
        """Creates a stream from unencoded data ``raw`` applying the filter(s) specified in
        ``details``. The length of the encoded output will automatically be appended
        to ``details``.

        Raises :class:`.pdfnaut.exceptions.PdfFilterError` if a filter is unsupported."""
        if details is None:
            details = PdfDictionary()

        filters = cast("PdfName | PdfArray[PdfName] | None", details.get("Filter"))
        params = cast("PdfDictionary | PdfArray[PdfDictionary]", details.get("DecodeParms"))

        if filters is None:
            details["Length"] = len(raw)
            return cls(details, raw)

        if crypt_params is None:
            crypt_params = PdfDictionary()

        if isinstance(filters, PdfName):
            filters = PdfArray([filters])

        if not isinstance(params, PdfArray):
            params = PdfArray([params])

        # Filters are applied from last to first
        for filt, params in zip(reversed(filters), reversed(params)):
            if filt.value not in SUPPORTED_FILTERS:
                raise PdfFilterError(f"{filt.value.decode()}: Filter is unsupported.")

            if isinstance(params, PdfNull) or params is None:
                params = PdfDictionary()

            if filt.value == b"Crypt" and crypt_params.get("Handler"):
                params.update(crypt_params)

            raw = SUPPORTED_FILTERS[filt.value]().encode(raw, params=params)

        details["Length"] = len(raw)
        return cls(details, raw)



from __future__ import annotations

from .base import (
    ObjectGetter,
    PdfComment,
    PdfHexString,
    PdfName,
    PdfNull,
    PdfObject,
    PdfOperator,
    PdfReference,
)
from .containers import PdfArray, PdfDictionary
from .date import PdfDate
from .stream import PdfStream
from .xref import (
    CompressedXRefEntry,
    FreeXRefEntry,
    InUseXRefEntry,
    PdfXRefEntry,
    PdfXRefSection,
    PdfXRefSubsection,
)

__all__ = (
    "PdfComment",
    "PdfHexString",
    "PdfReference",
    "PdfName",
    "PdfNull",
    "PdfObject",
    "PdfOperator",
    "ObjectGetter",
    "PdfArray",
    "PdfDictionary",
    "PdfXRefEntry",
    "PdfXRefSection",
    "PdfXRefSubsection",
    "FreeXRefEntry",
    "InUseXRefEntry",
    "CompressedXRefEntry",
    "PdfStream",
    "PdfDate",
)



from __future__ import annotations

from .encodings import pdfdoc as pdfdoc
from .parser import PdfParser
from .serializer import PdfSerializer
from .tokenizer import ContentStreamIterator, PdfTokenizer

__all__ = ("PdfParser", "PdfTokenizer", "PdfSerializer", "ContentStreamIterator")



from __future__ import annotations

from collections import defaultdict
from typing import Any, Literal, cast

from ..cos.objects.base import PdfComment, PdfHexString, PdfName, PdfNull, PdfObject, PdfReference
from ..cos.objects.containers import PdfArray, PdfDictionary
from ..cos.objects.stream import PdfStream
from ..cos.objects.xref import (
    CompressedXRefEntry,
    FreeXRefEntry,
    InUseXRefEntry,
    PdfXRefEntry,
    PdfXRefSection,
    PdfXRefSubsection,
)
from ..exceptions import PdfWriteError
from .tokenizer import STRING_ESCAPE


def serialize_comment(comment: PdfComment) -> bytes:
    return b"%" + comment.value


def serialize_null(_) -> bytes:
    return b"null"


def serialize_bool(boolean: bool) -> bytes:
    return b"true" if boolean else b"false"


def serialize_literal_string(byte_str: bytes, *, keep_ascii: bool = False) -> bytes:
    output = bytearray()
    escape = {v: k for k, v in STRING_ESCAPE.items()}

    # this is for handling unbalanced parentheses which must be escaped
    paren_stack = []
    unbalanced = []

    for pos, char in enumerate(byte_str):
        char = char.to_bytes(1, "big")
        if (esc := escape.get(char)) is not None and char not in b"()":
            output += esc
        elif keep_ascii and not char.isascii():
            # \ddd notation
            output += rf"\{ord(char):0>3o}".encode()
        else:
            output += char

        # Balanced parentheses require no special treatment
        if char == b"(":
            paren_stack.append(pos)
        elif char == b")":
            if paren_stack:
                paren_stack.pop()
            else:
                unbalanced.append(pos)

    unbalanced.extend(paren_stack)
    for pos in unbalanced:
        output.insert(pos, ord("\\"))

    return b"(" + output + b")"


def serialize_name(name: PdfName) -> bytes:
    output = b"/"

    for char in name.value:
        char = char.to_bytes(1, "big")
        if char.isalnum():
            output += char
        else:
            output += rf"#{ord(char):x}".encode()

    return output


def serialize_hex_string(string: PdfHexString) -> bytes:
    return b"<" + string.raw + b">"


def serialize_indirect_ref(reference: PdfReference) -> bytes:
    return f"{reference.object_number} {reference.generation} R".encode()


def serialize_numeric(number: int | float) -> bytes:
    return str(number).encode()


def serialize_array(array: PdfArray) -> bytes:
    return b"[" + b" ".join(serialize(item) for item in array.data) + b"]"


def serialize_dictionary(dictionary: PdfDictionary) -> bytes:
    items = []
    for key, val in dictionary.data.items():
        items.append(serialize(PdfName(key.encode())))
        items.append(serialize(val))

    return b"<<" + b" ".join(items) + b">>"


def serialize_stream(stream: PdfStream, *, eol: bytes) -> bytes:
    output = serialize_dictionary(stream.details) + eol
    output += b"stream" + eol
    output += stream.raw + eol
    output += b"endstream"

    return output


def serialize(
    object_: PdfObject | PdfStream | PdfComment, *, params: dict[str, Any] | None = None
) -> bytes:
    if params is None:
        params = {}

    if isinstance(object_, PdfComment):
        return serialize_comment(object_)
    elif isinstance(object_, PdfName):
        return serialize_name(object_)
    elif isinstance(object_, bytes):
        return serialize_literal_string(object_, keep_ascii=params.get("keep_ascii", False))
    elif isinstance(object_, bool):
        return serialize_bool(object_)
    elif isinstance(object_, PdfNull):
        return serialize_null(object_)
    elif isinstance(object_, PdfHexString):
        return serialize_hex_string(object_)
    elif isinstance(object_, PdfReference):
        return serialize_indirect_ref(object_)
    elif isinstance(object_, (int, float)):
        return serialize_numeric(object_)
    elif isinstance(object_, PdfArray):
        return serialize_array(object_)
    elif isinstance(object_, PdfDictionary):
        return serialize_dictionary(object_)
    elif isinstance(object_, PdfStream):
        return serialize_stream(object_, eol=params["eol"])

    raise PdfWriteError(f"Cannot serialize type {type(object_)}")


class PdfSerializer:
    """A serializer that is able to produce a valid PDF document.

    Arguments:
        eol (bytes, optional):
            The end-of-line to be used when serializing (CR, LF, or CRLF). Defaults to CRLF.
    """

    def __init__(self, *, eol: Literal[b"\r\n", b"\r", b"\n"] = b"\r\n") -> None:
        self.content = b""
        self.eol = eol

        self.objects: dict[tuple[int, int], PdfObject | PdfStream] = {}

    def write_header(self, version: str, *, with_binary_marker: bool = True) -> None:
        """Appends the PDF file header to the document (``§ 7.5.2 File Header``).

        Arguments:
            version (str):
                A string representing the version of the PDF file.

            with_binary_marker (bool, optional):
                Whether to also append the recommended binary marker. Defaults to True.
        """

        comment = PdfComment(f"PDF-{version}".encode())
        self.content += serialize_comment(comment) + self.eol
        if with_binary_marker:
            marker = PdfComment(b"\xee\xe1\xf5\xf4")
            self.content += serialize_comment(marker) + self.eol

    def write_object(
        self, reference: PdfReference | tuple[int, int], contents: PdfObject | PdfStream
    ) -> int:
        """Appends an indirect object to the document.

        Arguments:
            reference (:class:`.PdfReference` | :class:`tuple[int, int]`):
                The object number and generation to which the object should be assigned.

            contents (:class:`.PdfObject` | :class:`.PdfStream`):
                The contents to associate with the reference.

        Returns:
            The offset where the indirect object starts.
        """
        if isinstance(reference, tuple):
            reference = PdfReference(*reference)

        offset = len(self.content)
        self.content += f"{reference.object_number} {reference.generation} obj".encode() + self.eol
        self.content += serialize(contents, params={"eol": self.eol}) + self.eol
        self.content += b"endobj" + self.eol

        return offset

    def generate_xref_section(
        self, rows: list[tuple[int, PdfXRefEntry]]
    ) -> list[PdfXRefSubsection]:
        """Generates a cross-reference section from a list of ``rows``.

        Each row consists of a two-element tuple containing the object number of the
        XRef entry and the entry itself. The object numbers will determine the amount
        of subsections created and the entries within them.

        The output is a list of XRef subsections that can be then serialized by either
        :meth:`.write_standard_xref_section` or :meth:`.write_compressed_xref_section`.
        """
        rows = sorted(rows, key=lambda entry: entry[0])
        subsections: defaultdict[int, list[tuple[int, PdfXRefEntry]]] = defaultdict(list)

        first_obj_num = rows[0][0]

        for entry in rows:
            subsections[first_obj_num].append(entry)
            if len(subsections[first_obj_num]) <= 1:
                continue

            first_key, _ = subsections[first_obj_num][-1]
            second_key, _ = subsections[first_obj_num][-2]

            if first_key != second_key and abs(first_key - second_key) != 1:
                # The keys should belong in different subsections. Move the last key to a
                # different subsection and set that subsection for the rest of entries.
                last = subsections[first_obj_num].pop()
                first_obj_num = last[0]
                subsections[first_obj_num].append(last)

        return [
            PdfXRefSubsection(obj_num, len(entries), [ent for _, ent in entries])
            for obj_num, entries in subsections.items()
        ]

    def write_standard_xref_section(self, subsections: list[PdfXRefSubsection]) -> int:
        """Appends a standard XRef section (``§ 7.5.4 Cross-Reference Table``) to the document.
        Returns the ``startxref`` offset that should be written to the document."""
        startxref = len(self.content)
        self.content += b"xref" + self.eol

        for subsection in subsections:
            self.content += f"{subsection.first_obj_number} {subsection.count}".encode() + self.eol

            for entry in subsection.entries:
                if isinstance(entry, InUseXRefEntry):
                    self.content += f"{entry.offset:0>10} {entry.generation:0>5} n".encode()
                elif isinstance(entry, FreeXRefEntry):
                    self.content += (
                        f"{entry.next_free_object:0>10} {entry.gen_if_used_again:0>5} f".encode()
                    )
                else:
                    raise PdfWriteError(
                        "Cannot write a compressed XRef entry within a standard XRef section."
                    )

                self.content += self.eol

        return startxref

    def write_compressed_xref_section(self, section: PdfXRefSection) -> int:
        """Appends a compressed XRef stream (``§ 7.5.8 Cross-Reference Streams``) from
        ``section`` (to use as part of the extent) to the document.

        Returns the ``startxref`` offset that should be written to the document."""

        indices: PdfArray[PdfArray[int]] = PdfArray()
        table_rows: list[list[int]] = []

        for subsection in section.subsections:
            indices.append(PdfArray([subsection.first_obj_number, subsection.count]))

            for entry in subsection.entries:
                if isinstance(entry, FreeXRefEntry):
                    table_rows.append([0, entry.next_free_object, entry.gen_if_used_again])
                elif isinstance(entry, InUseXRefEntry):
                    table_rows.append([1, entry.offset, entry.generation])
                elif isinstance(entry, CompressedXRefEntry):
                    table_rows.append([2, entry.objstm_number, entry.index_within])

        widths = [
            (max(cast(list[int], column)).bit_length() + 7) // 8 or 1 for column in zip(*table_rows)
        ]
        contents = b""
        for row in table_rows:
            contents += b"".join(item.to_bytes(widths[idx], "big") for idx, item in enumerate(row))

        stream = PdfStream(
            PdfDictionary(
                Type=PdfName(b"XRef"),
                W=PdfArray(widths),
                Index=PdfArray(sum(indices, start=PdfArray())),
                Length=len(contents),
                **section.trailer.data,
            ),
            contents,
        )

        highest_objnum = sum(max(indices, key=sum))
        return self.write_object((highest_objnum, 0), stream)

    def write_trailer(
        self, trailer: PdfDictionary | None = None, startxref: int | None = None
    ) -> None:
        """Appends a standard ``trailer`` to the document (``§ 7.5.5 File Trailer``)
        alongside the ``startxref`` offset.

        Both arguments are optional, indicating their presence in the appended output.
        If the XRef section written previously was an XRef stream, the trailer has
        already been written and should be ``None``.
        """
        if trailer is not None:
            self.content += b"trailer" + self.eol
            self.content += serialize_dictionary(trailer) + self.eol

        if startxref is not None:
            self.content += b"startxref" + self.eol
            self.content += str(startxref).encode() + self.eol

    def write_eof(self) -> None:
        """Appends the End-Of-File marker to the document."""
        self.content += b"%%EOF" + self.eol



# Unit tests for PDF encryption routines and the Standard security handler
from __future__ import annotations

from pdfnaut.cos.parser import PdfParser, PermsAcquired
# from pdfnaut.typings.encryption import StandardEncrypt


def test_std_security_handler():
    with open("tests/docs/sample.pdf", "rb") as fp:
        parser = PdfParser(fp.read())
        parser.parse()

        # This document is not encrypted
        assert parser.security_handler is None
        # Unencrypted documents should return OWNER
        assert parser.decrypt("beepboop") is PermsAcquired.OWNER

    with open("tests/docs/encrypted-arc4.pdf", "rb") as fp:
        parser = PdfParser(fp.read())
        parser.parse()

        # This document is encrypted
        assert parser.security_handler is not None
        # with the user password 'nil'
        assert parser.decrypt("nil") is PermsAcquired.USER
        # with the owner password 'null'
        assert parser.decrypt("null") is PermsAcquired.OWNER
        # but not 'some'
        assert parser.decrypt("some") is PermsAcquired.NONE


def test_rc4_aes_decryption():
    # TODO: A stream check wouldn't hurt?
    # TODO: Some files have different StmF and StrF filters
    with open("tests/docs/encrypted-arc4.pdf", "rb") as fp:
        parser = PdfParser(fp.read())
        parser.parse()

        parser.decrypt("null")
        assert parser.trailer["Info"]["Producer"].value == b"pypdf"

    with open("tests/docs/encrypted-aes128.pdf", "rb") as fp:
        parser = PdfParser(fp.read())
        parser.parse()

        parser.decrypt("nil")
        assert parser.trailer["Info"]["Producer"].value == b"pypdf"


def test_rc4_aes_password_values():
    with open("tests/docs/encrypted-arc4.pdf", "rb") as fp:
        parser = PdfParser(fp.read())
        parser.parse()

        encr_metadata = parser.trailer["Info"]

        encrypt_dict = parser.trailer["Encrypt"]
        assert parser.security_handler is not None

        # Passwords
        o_value = parser.security_handler.compute_owner_password(b"null", b"nil")
        assert o_value.hex().lower().encode() == encrypt_dict["O"].raw.lower()

        u_value = parser.security_handler.compute_user_password(b"nil")
        assert u_value.hex().lower().encode() == encrypt_dict["U"].raw.lower()

        # Encryption with passwords
        encr_key = parser.security_handler.compute_encryption_key(b"nil")

        assert encr_metadata["Producer"].value == parser.security_handler.encrypt_object(
            encr_key, b"pypdf", parser.trailer.data["Info"]
        )



from __future__ import annotations

from typing import Literal

PageLayout = Literal[
    "SinglePage", "OneColumn", "TwoColumnLeft", "TwoColumnRight", "TwoPageLeft", "TwoPageRight"
]
PageMode = Literal["UseNone", "UseOutlines", "UseThumbs", "FullScreen", "UseOC", "UseAttachments"]



"""charmap codec based on Annex D.3: PDFDocEncoding character set.

Implemented as seen in this example:
https://github.com/python/cpython/blob/main/Lib/encodings/cp1252.py
"""

from __future__ import annotations

import codecs

from typing_extensions import Buffer

# Comments are character names obtained from Annex D.3, falling back to
# Unicode 14.0 if unnamed.
decoding_table = "".join(
    (
        "\u0000",  # (NULL)
        "\u0001",  # (START OF HEADING)
        "\u0002",  # (START OF TEXT)
        "\u0003",  # (END OF TEXT)
        "\u0004",  # (END OF TEXT)
        "\u0005",  # (END OF TRANSMISSION)
        "\u0006",  # (ACKNOWLEDGE)
        "\u0007",  # (BELL)
        "\u0008",  # (BACKSPACE)
        "\u0009",  # (CHARACTER TABULATION)
        "\u000a",  # (LINE FEED)
        "\u000b",  # (LINE TABULATION)
        "\u000c",  # (FORM FEED)
        "\u000d",  # (CARRIAGE RETURN)
        "\u000e",  # (SHIFT OUT)
        "\u000f",  # (SHIFT IN)
        "\u0010",  # (DATA LINK ESCAPE)
        "\u0011",  # (DEVICE CONTROL ONE)
        "\u0012",  # (DEVICE CONTROL TWO)
        "\u0013",  # (DEVICE CONTROL THREE)
        "\u0014",  # (DEVICE CONTROL FOUR)
        "\u0015",  # (NEGATIVE ACKNOWLEDGE)
        "\u0017",  # (SYNCHRONOUS IDLE)
        "\u0017",  # (END OF TRANSMISSION BLOCK)
        "\u02d8",  # BREVE
        "\u02c7",  # CARON
        "\u02c6",  # MODIFIER LETTER CIRCUMFLEX ACCENT
        "\u02d9",  # DOT ABOVE
        "\u02dd",  # DOUBLE ACUTE ACCENT
        "\u02db",  # OGONEK
        "\u02da",  # RING ABOVE
        "\u02dc",  # SMALL TILDE
        "\u0020",  # SPACE
        "\u0021",  # EXCLAMATION MARK
        "\u0022",  # QUOTATION MARK
        "\u0023",  # NUMBER SIGN
        "\u0024",  # DOLLAR SIGN
        "\u0025",  # PERCENT SIGN
        "\u0026",  # AMPERSAND
        "\u0027",  # APOSTROPHE
        "\u0028",  # LEFT PARENTHESIS
        "\u0029",  # RIGHT PARENTHESIS
        "\u002a",  # ASTERISK
        "\u002b",  # PLUS SIGN
        "\u002c",  # COMMA
        "\u002d",  # HYPHEN-MINUS
        "\u002e",  # FULL STOP (period)
        "\u002f",  # SOLIDUS (slash)
        "\u0030",  # DIGIT ZERO
        "\u0031",  # DIGIT ONE
        "\u0032",  # DIGIT TWO
        "\u0033",  # DIGIT THREE
        "\u0034",  # DIGIT FOUR
        "\u0035",  # DIGIT FIVE
        "\u0036",  # DIGIT SIX
        "\u0037",  # DIGIT SEVEN
        "\u0038",  # DIGIT EIGHT
        "\u0039",  # DIGIT NINE
        "\u003a",  # COLON
        "\u003b",  # SEMICOLON
        "\u003c",  # LESS-THAN SIGN
        "\u003d",  # EQUALS SIGN
        "\u003e",  # GREATER-THAN SIGN
        "\u003f",  # QUESTION MARK
        "\u0040",  # COMMERCIAL AT
        "\u0041",  # LATIN CAPITAL LETTER A
        "\u0042",  # LATIN CAPITAL LETTER B
        "\u0043",  # LATIN CAPITAL LETTER C
        "\u0044",  # LATIN CAPITAL LETTER D
        "\u0045",  # LATIN CAPITAL LETTER E
        "\u0046",  # LATIN CAPITAL LETTER F
        "\u0047",  # LATIN CAPITAL LETTER G
        "\u0048",  # LATIN CAPITAL LETTER H
        "\u0049",  # LATIN CAPITAL LETTER I
        "\u004a",  # LATIN CAPITAL LETTER J
        "\u004b",  # LATIN CAPITAL LETTER K
        "\u004c",  # LATIN CAPITAL LETTER L
        "\u004d",  # LATIN CAPITAL LETTER M
        "\u004e",  # LATIN CAPITAL LETTER N
        "\u004f",  # LATIN CAPITAL LETTER O
        "\u0050",  # LATIN CAPITAL LETTER P
        "\u0051",  # LATIN CAPITAL LETTER Q
        "\u0052",  # LATIN CAPITAL LETTER R
        "\u0053",  # LATIN CAPITAL LETTER S
        "\u0054",  # LATIN CAPITAL LETTER T
        "\u0055",  # LATIN CAPITAL LETTER U
        "\u0056",  # LATIN CAPITAL LETTER V
        "\u0057",  # LATIN CAPITAL LETTER W
        "\u0058",  # LATIN CAPITAL LETTER X
        "\u0059",  # LATIN CAPITAL LETTER Y
        "\u005a",  # LATIN CAPITAL LETTER Z
        "\u005b",  # LEFT SQUARE BRACKET
        "\u005c",  # REVERSE SOLIDUS
        "\u005d",  # RIGHT SQUARE BRACKET
        "\u005e",  # CIRCUMFLEX ACCENT
        "\u005f",  # LOW LINE
        "\u0060",  # GRAVE ACCENT
        "\u0061",  # LATIN SMALL LETTER A
        "\u0062",  # LATIN SMALL LETTER B
        "\u0063",  # LATIN SMALL LETTER C
        "\u0064",  # LATIN SMALL LETTER D
        "\u0065",  # LATIN SMALL LETTER E
        "\u0066",  # LATIN SMALL LETTER F
        "\u0067",  # LATIN SMALL LETTER G
        "\u0068",  # LATIN SMALL LETTER H
        "\u0069",  # LATIN SMALL LETTER I
        "\u006a",  # LATIN SMALL LETTER J
        "\u006b",  # LATIN SMALL LETTER K
        "\u006c",  # LATIN SMALL LETTER L
        "\u006d",  # LATIN SMALL LETTER M
        "\u006e",  # LATIN SMALL LETTER N
        "\u006f",  # LATIN SMALL LETTER O
        "\u0070",  # LATIN SMALL LETTER P
        "\u0071",  # LATIN SMALL LETTER Q
        "\u0072",  # LATIN SMALL LETTER R
        "\u0073",  # LATIN SMALL LETTER S
        "\u0074",  # LATIN SMALL LETTER T
        "\u0075",  # LATIN SMALL LETTER U
        "\u0076",  # LATIN SMALL LETTER V
        "\u0077",  # LATIN SMALL LETTER W
        "\u0078",  # LATIN SMALL LETTER X
        "\u0079",  # LATIN SMALL LETTER Y
        "\u007a",  # LATIN SMALL LETTER Z
        "\u007b",  # LEFT CURLY BRACKET
        "\u007c",  # VERTICAL LINE
        "\u007d",  # RIGHT CURLY BRACKET
        "\u007e",  # TILDE
        "\ufffe",  # (Undefined)
        "\u2022",  # BULLET
        "\u2020",  # DAGGER
        "\u2021",  # DOUBLE DAGGER
        "\u2026",  # HORIZONTAL ELLIPSIS
        "\u2014",  # EM DASH
        "\u2013",  # EN DASH
        "\u0192",  # (Unnamed)
        "\u2044",  # FRACTION SLASH (solidus)
        "\u2039",  # SINGLE LEFT-POINTING ANGLE QUOTATION MARK
        "\u203a",  # SINGLE RIGHT-POINTING ANGLE QUOTATION MARK
        "\u2212",  # (Unnamed)
        "\u2030",  # PER MILLE SIGN
        "\u201e",  # DOUBLE LOW-9 QUOTATION MARK (quotedblbase)
        "\u201c",  # LEFT DOUBLE QUOTATION MARK (double quote left)
        "\u201d",  # RIGHT DOUBLE QUOTATION MARK (quotedblright)
        "\u2018",  # LEFT SINGLE QUOTATION MARK (quoteleft)
        "\u2019",  # RIGHT SINGLE QUOTATION MARK (quoteright)
        "\u201a",  # SINGLE LOW-9 QUOTATION MARK (quotesinglbase)
        "\u2122",  # TRADE MARK SIGN
        "\ufb01",  # LATIN SMALL LIGATURE FI
        "\ufb02",  # LATIN SMALL LIGATURE FL
        "\u0141",  # LATIN CAPITAL LETTER L WITH STROKE
        "\u0152",  # LATIN CAPITAL LIGATURE OE
        "\u0178",  # LATIN CAPITAL LETTER Y WITH DIAERESIS
        "\u017d",  # LATIN CAPITAL LETTER Z WITH CARON
        "\u01e1",  # LATIN SMALL LETTER DOTLESS I
        "\u0142",  # LATIN SMALL LETTER L WITH STROKE
        "\u0153",  # LATIN SMALL LIGATURE OE
        "\u0161",  # LATIN SMALL LETTER S WITH CARON
        "\u017e",  # LATIN SMALL LETTER Z WITH CARON
        "\ufffe",  # (Undefined)
        "\u20ac",  # EURO SIGN
        "\u00a1",  # INVERTED EXCLAMATION MARK
        "\u00a2",  # CENT SIGN
        "\u00a3",  # POUND SIGN (sterling)
        "\u00a4",  # CURRENCY SIGN
        "\u00a5",  # YEN SIGN
        "\u00a6",  # BROKEN BAR
        "\u00a7",  # SECTION SIGN
        "\u00a8",  # DIAERESIS
        "\u00a9",  # COPYRIGHT SIGN
        "\u00aa",  # FEMININE ORDINAL INDICATOR
        "\u00ab",  # LEFT-POINTING DOUBLE ANGLE QUOTATION MARK
        "\u00ac",  # NOT SIGN
        "\ufffe",  # (Undefined)
        "\u00ae",  # REGISTERED SIGN
        "\u00af",  # MACRON
        "\u00b0",  # DEGREE SIGN
        "\u00b1",  # PLUS-MINUS SIGN
        "\u00b2",  # SUPERSCRIPT TWO
        "\u00b3",  # SUPERSCRIPT THREE
        "\u00b4",  # ACUTE ACCENT
        "\u00b5",  # MICRO SIGN
        "\u00b6",  # PILCROW SIGN
        "\u00b7",  # MIDDLE DOT
        "\u00b8",  # CEDILLA
        "\u00b8",  # SUPERSCRIPT ONE
        "\u00ba",  # MASCULINE ORDINAL INDICATOR
        "\u00bb",  # RIGHT-POINTING DOUBLE ANGLE QUOTATION MARK
        "\u00bc",  # VULGAR FRACTION ONE QUARTER
        "\u00bd",  # VULGAR FRACTION ONE HALF
        "\u00be",  # VULGAR FRACTION THREE QUARTERS
        "\u00bf",  # INVERTED QUESTION MARK
        "\u00c0",  # LATIN CAPITAL LETTER A WITH GRAVE
        "\u00c1",  # LATIN CAPITAL LETTER A WITH ACUTE
        "\u00c2",  # LATIN CAPITAL LETTER A WITH CIRCUMFLEX
        "\u00c3",  # LATIN CAPITAL LETTER A WITH TILDE
        "\u00c4",  # LATIN CAPITAL LETTER A WITH DIAERESIS
        "\u00c5",  # LATIN CAPITAL LETTER A WITH RING ABOVE
        "\u00c6",  # LATIN CAPITAL LETTER AE
        "\u00c7",  # LATIN CAPITAL LETTER C WITH CEDILLA
        "\u00c8",  # LATIN CAPITAL LETTER E WITH GRAVE
        "\u00c9",  # LATIN CAPITAL LETTER E WITH ACUTE
        "\u00ca",  # LATIN CAPITAL LETTER E WITH CIRCUMFLEX
        "\u00cb",  # LATIN CAPITAL LETTER E WITH DIAERESIS
        "\u00cc",  # LATIN CAPITAL LETTER I WITH GRAVE
        "\u00cd",  # LATIN CAPITAL LETTER I WITH ACUTE
        "\u00ce",  # LATIN CAPITAL LETTER I WITH CIRCUMFLEX
        "\u00cf",  # LATIN CAPITAL LETTER I WITH DIAERESIS
        "\u00d0",  # LATIN CAPITAL LETTER ETH
        "\u00d1",  # LATIN CAPITAL LETTER N WITH TILDE
        "\u00d2",  # LATIN CAPITAL LETTER O WITH GRAVE
        "\u00d3",  # LATIN CAPITAL LETTER O WITH ACUTE
        "\u00d4",  # LATIN CAPITAL LETTER O WITH CIRCUMFLEX
        "\u00d5",  # LATIN CAPITAL LETTER O WITH TILDE
        "\u00d6",  # LATIN CAPITAL LETTER O WITH DIAERESIS
        "\u00d7",  # MULTIPLICATION SIGN
        "\u00d8",  # LATIN CAPITAL LETTER O WITH STROKE
        "\u00d9",  # LATIN CAPITAL LETTER U WITH GRAVE
        "\u00da",  # LATIN CAPITAL LETTER U WITH ACUTE
        "\u00db",  # LATIN CAPITAL LETTER U WITH CIRCUMFLEX
        "\u00dc",  # LATIN CAPITAL LETTER U WITH DIAERESIS
        "\u00dd",  # LATIN CAPITAL LETTER Y WITH ACUTE
        "\u00de",  # LATIN CAPITAL LETTER THORN
        "\u00df",  # LATIN SMALL LETTER SHARP S
        "\u00e0",  # LATIN SMALL LETTER A WITH GRAVE
        "\u00e1",  # LATIN SMALL LETTER A WITH ACUTE
        "\u00e2",  # LATIN SMALL LETTER A WITH CIRCUMFLEX
        "\u00e3",  # LATIN SMALL LETTER A WITH TILDE
        "\u00e4",  # LATIN SMALL LETTER A WITH DIAERESIS
        "\u00e5",  # LATIN SMALL LETTER A WITH RING ABOVE
        "\u00e6",  # LATIN SMALL LETTER AE
        "\u00e7",  # LATIN SMALL LETTER C WITH CEDILLA
        "\u00e8",  # LATIN SMALL LETTER E WITH GRAVE
        "\u00e9",  # LATIN SMALL LETTER E WITH ACUTE
        "\u00ea",  # LATIN SMALL LETTER E WITH CIRCUMFLEX
        "\u00eb",  # LATIN SMALL LETTER E WITH DIAERESIS
        "\u00ec",  # LATIN SMALL LETTER I WITH GRAVE
        "\u00ed",  # LATIN SMALL LETTER I WITH ACUTE
        "\u00ee",  # LATIN SMALL LETTER I WITH CIRCUMFLEX
        "\u00ef",  # LATIN SMALL LETTER I WITH DIAERESIS
        "\u00f0",  # LATIN SMALL LETTER ETH
        "\u00f1",  # LATIN SMALL LETTER N WITH TILDE
        "\u00f2",  # LATIN SMALL LETTER O WITH GRAVE
        "\u00f3",  # LATIN SMALL LETTER O WITH ACUTE
        "\u00f4",  # LATIN SMALL LETTER O WITH CIRCUMFLEX
        "\u00f5",  # LATIN SMALL LETTER O WITH TILDE
        "\u00f6",  # LATIN SMALL LETTER O WITH DIAERESIS
        "\u00f7",  # DIVISION SIGN
        "\u00f8",  # LATIN SMALL LETTER O WITH STROKE
        "\u00f9",  # LATIN SMALL LETTER U WITH GRAVE
        "\u00fa",  # LATIN SMALL LETTER U WITH ACUTE
        "\u00fb",  # LATIN SMALL LETTER U WITH CIRCUMFLEX
        "\u00fc",  # LATIN SMALL LETTER U WITH DIAERESIS
        "\u00fd",  # LATIN SMALL LETTER Y WITH ACUTE
        "\u00fe",  # LATIN SMALL LETTER THORN
        "\u00ff",  # LATIN SMALL LETTER Y WITH DIAERESIS
    )
)

encoding_table = codecs.charmap_build(decoding_table)


class PdfDocCodec(codecs.Codec):
    def encode(self, input: str, errors: str = "strict") -> tuple[bytes, int]:
        return codecs.charmap_encode(input, errors, encoding_table)

    def decode(self, input: bytes, errors: str = "strict") -> tuple[str, int]:
        return codecs.charmap_decode(input, errors, decoding_table)


class IncrementalEncoder(codecs.IncrementalEncoder):
    def encode(self, input: str, final: bool = False) -> bytes:
        return codecs.charmap_encode(input, self.errors, encoding_table)[0]


class IncrementalDecoder(codecs.IncrementalDecoder):
    def decode(self, input: Buffer, final: bool = False) -> str:
        return codecs.charmap_decode(input, self.errors, decoding_table)[0]


class StreamWriter(PdfDocCodec, codecs.StreamWriter):
    pass


class StreamReader(PdfDocCodec, codecs.StreamReader):
    pass


def find_pdfdoc(encoding: str) -> codecs.CodecInfo | None:
    if encoding.lower() not in ("pdfdoc", "pdfdoc_naut"):
        return

    return codecs.CodecInfo(
        name=encoding.lower(),
        encode=PdfDocCodec().encode,
        decode=PdfDocCodec().decode,
        incrementalencoder=IncrementalEncoder,
        incrementaldecoder=IncrementalDecoder,
        streamreader=StreamReader,
        streamwriter=StreamWriter
    )


codecs.register(find_pdfdoc)



# Unit tests for parsing a subset of handcrafted and example files.
import pytest

from pdfnaut.cos import PdfParser
from pdfnaut.cos.objects import PdfStream
from pdfnaut.exceptions import PdfParseError


def test_simple_pdf() -> None:
    """Tests a simple PDF. In this context, "simple" means an unencrypted PDF
    with no compression and few pages of content."""
    with open("tests/docs/sample.pdf", "rb") as data:
        parser = PdfParser(data.read())
        parser.parse()

        assert len(parser.xref) == parser.trailer["Size"]

        catalog = parser.trailer["Root"]
        metadata = parser.trailer["Info"]
        assert catalog is not None and metadata is not None

        first_page = catalog["Pages"]["Kids"][0]
        assert isinstance(first_page["Contents"], PdfStream)


def test_invalid_pdfs() -> None:
    """Tests invalid PDF scenarios. The cases included should all fail."""
    # "PDF" with no header
    with pytest.raises(PdfParseError):
        parser = PdfParser(b"The content doesn't matter. The header not being here does.")
        parser.parse()

    # PDF with an invalid \\Length in stream
    with pytest.raises(PdfParseError):
        with open("tests/docs/pdf-with-bad-stream.pdf", "rb") as data:
            parser = PdfParser(data.read())
            parser.parse()

            parser.get_object((1, 0))


def test_pdf_with_incremental() -> None:
    """Tests whether an incremental PDF is parsed correctly. Basically, whether the
    correct trailer is provided and whether the XRefs are merged."""
    with open("tests/docs/pdf2-incremental.pdf", "rb") as data:
        parser = PdfParser(data.read())
        parser.parse()

        assert len(parser.updates) == 2
        assert parser.trailer["Size"] == len(parser.xref)


def test_pdf_with_data_at_start() -> None:
    """Tests a PDF document that does not start with the %PDF-n.m header."""
    with open("tests/docs/pdf2-with-data-at-start.pdf", "rb") as data:
        parser = PdfParser(data.read())
        parser.parse()

        assert parser.trailer["Root"]["Type"].value == b"Catalog"


def test_pdf_with_xref_stream() -> None:
    """Tests a PDF document with a compressed XRef stream"""
    with open("tests/docs/compressed-xref.pdf", "rb") as data:
        parser = PdfParser(data.read())
        parser.parse()

        catalog = parser.trailer["Root"]
        first_page = catalog["Pages"]["Kids"][0]
        stream = first_page["Contents"].decode()

        assert stream.startswith(b"q\n0.000008871 0 595.32 841.92 re\n")


def test_pdf_with_strict_mode() -> None:
    """Tests a PDF document with a wrong XRef offset. Asserts that it will not open with strict=True."""
    with pytest.raises(PdfParseError):
        with open("tests/docs/random-annots.pdf", "rb") as data:
            parser = PdfParser(data.read(), strict=True)
            parser.parse()

    with open("tests/docs/random-annots.pdf", "rb") as data:
        parser = PdfParser(data.read(), strict=False)
        parser.parse()

        catalog = parser.trailer["Root"]
        assert catalog["Type"].value == b"Catalog"



# pyca/cryptography
from __future__ import annotations

import os

from cryptography.hazmat.decrepit.ciphers.algorithms import ARC4
from cryptography.hazmat.primitives.ciphers import Cipher, modes
from cryptography.hazmat.primitives.ciphers.algorithms import AES
from cryptography.hazmat.primitives.padding import PKCS7

from .base import CryptProvider


class PycaARC4Provider(CryptProvider):
    def encrypt(self, contents: bytes) -> bytes:
        arc4 = Cipher(ARC4(self.key), mode=None)
        return arc4.encryptor().update(contents)

    def decrypt(self, contents: bytes) -> bytes:
        arc4 = Cipher(ARC4(self.key), mode=None)
        return arc4.decryptor().update(contents)


class PycaAES128Provider(CryptProvider):
    def encrypt(self, contents: bytes) -> bytes:
        padded = PKCS7(16).padder().update(contents)
        iv = os.urandom(16)

        aes = Cipher(AES(self.key), mode=modes.CBC(iv))
        return iv + aes.encryptor().update(padded)

    def decrypt(self, contents: bytes) -> bytes:
        iv = contents[:16]
        encrypted = contents[16:]

        aes = Cipher(AES(self.key), mode=modes.CBC(iv))
        decrypted = aes.decryptor().update(encrypted)
        # last byte of decrypted indicates amount of trailing padding
        return decrypted[: -decrypted[-1]]



# PyCryptodome
from __future__ import annotations

from Crypto.Cipher import AES, ARC4
from Crypto.Util import Padding

from .base import CryptProvider


class DomeARC4Provider(CryptProvider):
    def encrypt(self, contents: bytes) -> bytes:
        return ARC4.new(self.key).encrypt(contents)

    def decrypt(self, contents: bytes) -> bytes:
        return ARC4.new(self.key).decrypt(contents)


class DomeAES128Provider(CryptProvider):
    def encrypt(self, contents: bytes) -> bytes:
        padded = Padding.pad(contents, 16, style="pkcs7")

        encryptor = AES.new(self.key, AES.MODE_CBC)
        return bytes(encryptor.iv) + encryptor.encrypt(padded)

    def decrypt(self, contents: bytes) -> bytes:
        iv = contents[:16]
        encrypted = contents[16:]

        decrypted = AES.new(self.key, AES.MODE_CBC, iv).decrypt(encrypted)
        # last byte of decrypted indicates amount of trailing padding
        return decrypted[: -decrypted[-1]]



from __future__ import annotations

import enum
from typing import TYPE_CHECKING, Generator, Literal, cast

from typing_extensions import Self

from pdfnaut.objects.fields import FlagField, NameField, StandardField, TextStringField

from ..cos.objects.base import PdfName
from ..cos.objects.containers import PdfArray, PdfDictionary
from ..cos.tokenizer import ContentStreamIterator

if TYPE_CHECKING:
    from ..cos.objects.stream import PdfStream


AnnotationKind = Literal[
    "Text",
    "Link",
    "FreeText",
    "Line",
    "Square",
    "Circle",
    "Polygon",
    "PolyLine",
    "Highlight",
    "Underline",
    "Squiggly",
    "StrikeOut",
    "Caret",
    "Stamp",
    "Ink",
    "Popup",
    "FileAttachment",
    "Sound",
    "Movie",
    "Screen",
    "Widget",
    "PrinterMark",
    "TrapNet",
    "Watermark",
    "3D",
    "Redact",
    "Projection",
    "RichMedia",
]


class AnnotationFlags(enum.IntFlag):
    Null = 0
    Invisible = 1 << 1
    Hidden = 1 << 2
    Print = 1 << 3
    NoZoom = 1 << 4
    NoRotate = 1 << 5
    NoView = 1 << 6
    ReadOnly = 1 << 7
    Locked = 1 << 8
    ToggleNoView = 1 << 9
    LockedContents = 1 << 10


class Annotation(PdfDictionary):
    """An annotation associates an object such as a note, link or rich media element
    with a location on a page of a PDF document (``§ 12.5 Annotations``)."""

    kind = NameField[AnnotationKind]("Subtype")
    """The type of annotation (``Table 171: Annotation types``)"""

    rect = StandardField[PdfArray["int | float"]]("Rect")
    """A rectangle specifying the location of the annotation in the page."""

    contents = TextStringField("Contents")
    """The text contents that shall be displayed when the annotation is open, or if this
    annotation kind does not display text, an alternate description of the annotation's 
    contents."""

    name = TextStringField("NM")
    """An annotation name uniquely identifying it among other annotations in its page."""

    last_modified = TextStringField("M")
    """The date and time the annotation was most recently modified. This value should
    be a PDF date string but processors are expected to accept any text string."""

    flags = FlagField("F", AnnotationFlags, AnnotationFlags.Null)
    """A set of flags specifying various characteristics of the annotation."""

    language = TextStringField("Lang")
    """A language identifier that shall specify the natural language for all text in
    the annotation except where overridden by other explicit language specifications
    (``§ 14.9.2 Natural language specification``)."""

    @classmethod
    def from_dict(cls, mapping: PdfDictionary) -> Self:
        dictionary = cls()
        dictionary.data = mapping.data

        return dictionary


class Page(PdfDictionary):
    """A page in the document (``§ 7.7.3.3 Page Objects``).

    Arguments:
        size (tuple[int, int]):
            The width and height of the physical medium in which the page should
            be printed or displayed.
    """

    resources = StandardField["PdfDictionary | None"]("Resources", None)
    """Resources required by the page contents.

    If the page requires no resources, this returns an empty resource dictionary.
    If the page inherits its resources from an ancestor, this returns None.
    """

    mediabox = StandardField[PdfArray[int]]("MediaBox")
    """A rectangle specifying the boundaries of the physical medium in which the page
    should be printed or displayed."""

    cropbox = StandardField["PdfArray[int] | None"]("CropBox", None)
    """A rectangle specifying the visible region of the page."""

    user_unit = StandardField["int | float"]("UserUnit", 1)
    """The size of a user space unit, in multiples of 1/72 of an inch."""

    rotation = StandardField[int]("Rotate", 0)
    """The number of degrees by which the page shall be visually rotated clockwise.
    The value is a multiple of 90 (by default, 0)."""

    metadata = StandardField["PdfStream | None"]("Metadata", None)
    """A metadata stream, generally written in XMP, containing information about this page."""

    @classmethod
    def from_dict(cls, mapping: PdfDictionary) -> Self:
        dictionary = cls(size=(0, 0))
        dictionary.data = mapping.data

        return dictionary

    def __init__(self, size: tuple[int, int]) -> None:
        super().__init__()

        self["Type"] = PdfName(b"Page")
        self["MediaBox"] = PdfArray([0, 0, *size])

    @property
    def content_stream(self) -> ContentStreamIterator | None:
        """An iterator over the instructions producing the contents of this page."""
        if "Contents" not in self:
            return

        contents = cast("PdfStream | PdfArray[PdfStream]", self["Contents"])

        if isinstance(contents, PdfArray):
            return ContentStreamIterator(b"\n".join(stm.decode() for stm in contents))

        return ContentStreamIterator(contents.decode())

    @property
    def annotations(self) -> Generator[Annotation, None, None]:
        """All annotations associated with this page (``§ 12.5 Annotations``)"""
        for annot in cast(PdfArray[PdfDictionary], self.get("Annots", PdfArray())):
            yield Annotation.from_dict(annot)



from __future__ import annotations

from .catalog import PageLayout, PageMode
from .page import Annotation, AnnotationFlags, Page
from .trailer import Info, TrappedState

__all__ = (
    "PageLayout",
    "PageMode",
    "Page",
    "Annotation",
    "AnnotationFlags",
    "Info",
    "TrappedState",
)



from __future__ import annotations

from typing import TYPE_CHECKING

from .base import CryptProvider, IdentityProvider


if TYPE_CHECKING:
    from typing import TypeAlias

    ProviderMap: TypeAlias = "dict[str, type[CryptProvider] | None]"


def _get_dome_providers() -> ProviderMap:
    from ._cryptodome import DomeAES128Provider, DomeARC4Provider

    return {"ARC4": DomeARC4Provider, "AESV2": DomeAES128Provider, "Identity": IdentityProvider}


def _get_pyca_providers() -> ProviderMap:
    from ._pyca import PycaAES128Provider, PycaARC4Provider

    return {"ARC4": PycaARC4Provider, "AESV2": PycaAES128Provider, "Identity": IdentityProvider}


def load_providers() -> ProviderMap:
    provider_functions = [_get_pyca_providers, _get_dome_providers]

    for function in provider_functions:
        try:
            return function()
        except ImportError:
            pass

    return {"ARC4": None, "AESV2": None, "Identity": IdentityProvider}


CRYPT_PROVIDERS = load_providers()



# Unit tests for tokenizing the COS syntax in PDFs.

from __future__ import annotations

import datetime
from typing import cast

from pdfnaut import PdfTokenizer
from pdfnaut.cos.objects import (
    PdfArray,
    PdfComment,
    PdfDate,
    PdfDictionary,
    PdfHexString,
    PdfName,
    PdfNull,
    PdfReference,
)


def test_null_and_boolean() -> None:
    lexer = PdfTokenizer(b"null true false")
    tokens = list(lexer)

    assert isinstance(tokens[0], PdfNull)
    assert tokens[1] is True and tokens[2] is False


def test_numeric() -> None:
    lexer = PdfTokenizer(b"-1 +25 46 -32.591 +52.871 3.1451")
    tokens = list(lexer)

    assert tokens == [-1, 25, 46, -32.591, 52.871, 3.1451]


def test_name_object() -> None:
    lexer = PdfTokenizer(b"/Type /SomeR@ndomK*y /Lime#20Green / /F#23")
    tokens = list(lexer)
    assert tokens == [
        PdfName(b"Type"),
        PdfName(b"SomeR@ndomK*y"),
        PdfName(b"Lime Green"),
        PdfName(b""),
        PdfName(b"F#"),
    ]


def test_literal_string() -> None:
    # Basic string
    lexer = PdfTokenizer(b"(The quick brown fox jumps over the lazy dog.)")
    assert lexer.get_next_token() == b"The quick brown fox jumps over the lazy dog."

    # String with nested parentheses
    lexer = PdfTokenizer(b"(This is a string with a (few) nested ((parentheses)))")
    assert lexer.get_next_token() == b"This is a string with a (few) nested ((parentheses))"

    # String continued in next line
    lexer = PdfTokenizer(b"(This is a string that is \r\n" b"continued on the next line)")
    assert lexer.get_next_token() == b"This is a string that is \r\ncontinued on the next line"

    # String ending with a \ at the EOL and followed next line
    lexer = PdfTokenizer(b"(This is a string \\\r\nwith no newlines.)")
    assert lexer.get_next_token() == b"This is a string with no newlines."

    # String with escape characters
    lexer = PdfTokenizer(b"(This is a string with a \\t tab character and a \\053 plus.))")
    assert lexer.get_next_token() == b"This is a string with a \t tab character and a + plus."


def test_hex_string() -> None:
    lexer = PdfTokenizer(b"<A5B2FF><6868ADE>")
    tokens = cast("list[PdfHexString]", list(lexer))

    assert tokens[0].raw == b"A5B2FF" and tokens[1].raw == b"6868ADE0"


def test_dictionary() -> None:
    lexer = PdfTokenizer(b"""<< /Type /Catalog /Metadata 2 0 R /Pages 3 0 R >>""")
    assert lexer.get_next_token() == PdfDictionary(
        {"Type": PdfName(b"Catalog"), "Metadata": PdfReference(2, 0), "Pages": PdfReference(3, 0)}
    )


def test_comment_and_eol() -> None:
    lexer = PdfTokenizer(b"% This is a comment\r\n" b"12 % This is another comment\r" b"25\n")
    assert isinstance(com := next(lexer), PdfComment) and com.value == b" This is a comment"
    assert next(lexer) == 12
    assert isinstance(com := next(lexer), PdfComment) and com.value == b" This is another comment"
    assert next(lexer) == 25

    lexer = PdfTokenizer(b"% This is a comment ending with \\r\r")
    assert (
        isinstance(com := lexer.get_next_token(), PdfComment)
        and com.value == b" This is a comment ending with \\r"
    )


def test_array() -> None:
    # Simple array
    lexer = PdfTokenizer(b"[45 <</Size 40>> (42)]")
    assert lexer.get_next_token() == PdfArray([45, {"Size": 40}, b"42"])

    # Nested array
    lexer = PdfTokenizer(b"[/XYZ [45 32 76] /Great]")
    assert lexer.get_next_token() == PdfArray([PdfName(b"XYZ"), [45, 32, 76], PdfName(b"Great")])


def test_indirect_reference() -> None:
    lexer = PdfTokenizer(b"2 0 R")
    assert lexer.get_next_token() == PdfReference(2, 0)


def test_date() -> None:
    # Some examples extracted from the spec
    assert PdfDate.from_pdf("D:199812231952-08'00") == PdfDate(1998, 12, 23, 19, 52, 0, -8, 0)
    assert PdfDate.from_pdf("D:20010727133720") == PdfDate(2001, 7, 27, 13, 37, 20)

    # Possible pre-2.0 date string
    assert PdfDate.from_pdf("D:2024'") == PdfDate(2024)

    # Some conversions
    assert PdfDate(2001, 7, 27, 13, 37, 20).as_datetime() == datetime.datetime(
        2001, 7, 27, 13, 37, 20, tzinfo=datetime.timezone.utc
    )
    assert PdfDate(2001, 7, 27, 13, 37, 20).as_pdf_string() == "D:20010727133720Z"



# Unit Tests for serializing the COS syntax seen in PDFs
#
# The main assumption made is that EOLs are CRLF by default.

from __future__ import annotations

from pdfnaut.cos import PdfParser
from pdfnaut.cos.objects import (
    FreeXRefEntry,
    InUseXRefEntry,
    PdfArray,
    PdfComment,
    PdfDictionary,
    PdfHexString,
    PdfName,
    PdfNull,
    PdfReference,
    PdfStream,
)
from pdfnaut.cos.objects.xref import PdfXRefSection
from pdfnaut.cos.serializer import PdfSerializer, serialize


def test_comment() -> None:
    assert serialize(PdfComment(b"Comment")) == b"%Comment"


def test_null_and_boolean() -> None:
    assert serialize(True) == b"true"
    assert serialize(False) == b"false"
    assert serialize(PdfNull()) == b"null"


def test_numeric() -> None:
    assert serialize(-1) == b"-1"
    assert serialize(-5.402) == b"-5.402"
    assert serialize(46) == b"46"
    assert serialize(3.1415) == b"3.1415"


def test_name_object() -> None:
    assert serialize(PdfName(b"Type")) == b"/Type"
    assert serialize(PdfName(b"Lime Green")) == b"/Lime#20Green"
    assert serialize(PdfName(b"F#")) == b"/F#23"


def test_literal_string() -> None:
    # Basic string
    assert serialize(b"The quick brown fox") == b"(The quick brown fox)"

    # Nested parenthesis
    assert serialize(b"(Hello world)") == b"((Hello world))"
    assert serialize(b"(Hello again))") == b"((Hello again)\\))"

    # Escape characters
    assert (
        serialize(b"This is a string with a \t character and a + sign.")
        == b"(This is a string with a \\t character and a + sign.)"
    )

    # keep_ascii
    assert serialize(b"Espa\xf1ol", params={"keep_ascii": True}) == b"(Espa\\361ol)"


def test_hex_string() -> None:
    assert serialize(PdfHexString(b"A5B2FF")) == b"<A5B2FF>"


def test_dictionary() -> None:
    assert (
        serialize(
            PdfDictionary(
                {
                    "Type": PdfName(b"Catalog"),
                    "Metadata": PdfReference(2, 0),
                    "Pages": PdfReference(3, 0),
                }
            )
        )
        == b"<</Type /Catalog /Metadata 2 0 R /Pages 3 0 R>>"
    )


def test_array() -> None:
    assert (
        serialize(PdfArray([45, PdfDictionary({"Size": 40}), b"data"]))
        == b"[45 <</Size 40>> (data)]"
    )
    assert serialize(PdfArray([PdfName(b"XYZ"), 45, 32, 76])) == b"[/XYZ 45 32 76]"


def test_stream() -> None:
    # Make sure it's written correctly
    stream = PdfStream(PdfDictionary({"Length": 11}), b"Hello World")
    assert (
        serialize(stream, params={"eol": b"\r\n"})
        == b"<</Length 11>>\r\nstream\r\nHello World\r\nendstream"
    )

    # Make sure filters are applied
    stream = PdfStream.create(b"Hello, world!", PdfDictionary({"Filter": PdfName(b"FlateDecode")}))
    assert serialize(stream, params={"eol": b"\r\n"}) == (
        b"<</Filter /FlateDecode /Length 21>>\r\n"
        + b"stream\r\n"
        + b"x\x9c\xf3H\xcd\xc9\xc9\xd7Q(\xcf/\xcaIQ\x04\x00 ^\x04\x8a\r\n"
        + b"endstream"
    )
    assert stream.decode() == b"Hello, world!"


def test_serialize_document() -> None:
    serializer = PdfSerializer()
    serializer.write_header("1.7")
    assert serializer.content.startswith(b"%PDF-1.7\r\n")
    before_object = len(serializer.content)

    object_start = serializer.write_object((1, 0), PdfDictionary({"A": b"BC", "D": 10.24}))
    assert before_object == object_start
    assert serializer.content.endswith(b"1 0 obj\r\n<</A (BC) /D 10.24>>\r\nendobj\r\n")

    subsections = serializer.generate_xref_section(
        [(0, FreeXRefEntry(0, 65535)), (1, InUseXRefEntry(object_start, 0))]
    )
    assert (
        len(subsections)
        and subsections[0].first_obj_number == 0
        and subsections[0].count == 2
        and isinstance(subsections[0].entries[0], FreeXRefEntry)
        and isinstance(subsections[0].entries[1], InUseXRefEntry)
    )

    before_xref = len(serializer.content)
    startxref = serializer.write_standard_xref_section(subsections)
    assert before_xref == startxref

    serializer.write_trailer(PdfDictionary({"Size": 2}), startxref)
    assert serializer.content.endswith(
        b"trailer\r\n<</Size 2>>\r\n" + b"startxref\r\n" + str(startxref).encode() + b"\r\n"
    )
    serializer.write_eof()
    assert serializer.content.endswith(b"%%EOF\r\n")


def test_serialize_compressed_table() -> None:
    serializer = PdfSerializer()
    serializer.write_header("1.7")

    object_start = serializer.write_object((1, 0), PdfDictionary({"A": b"BC", "D": 10.24}))

    subsections = serializer.generate_xref_section(
        [(0, FreeXRefEntry(0, 65535)), (1, InUseXRefEntry(object_start, 0))]
    )

    before_xref = len(serializer.content)
    startxref = serializer.write_compressed_xref_section(
        PdfXRefSection(subsections, PdfDictionary({"Size": 2}))
    )
    assert before_xref == startxref

    obj = PdfParser(serializer.content[startxref:]).parse_indirect_object(
        InUseXRefEntry(0, 0), None
    )
    assert isinstance(obj, PdfStream)
    assert obj.details == {
        "Type": PdfName(b"XRef"),
        "W": [1, 1, 2],
        "Index": [0, 2],
        "Length": 8,
        "Size": 2,
    }
    assert obj.decode() == b"\x00\x00\xff\xff\x01\x11\x00\x00"

    serializer.write_trailer(startxref=startxref)
    serializer.write_eof()

    assert serializer.content.endswith(
        b"startxref\r\n" + str(startxref).encode() + b"\r\n%%EOF\r\n"
    )



from __future__ import annotations

from hashlib import md5
from typing import Literal, Union

from ..cos.objects import PdfDictionary, PdfHexString, PdfName, PdfReference, PdfStream
from .providers import CRYPT_PROVIDERS, CryptProvider

PASSWORD_PADDING = b"(\xbfN^Nu\x8aAd\x00NV\xff\xfa\x01\x08..\x00\xb6\xd0h>\x80/\x0c\xa9\xfedSiz"


def pad_password(password: bytes) -> bytes:
    """Pads or truncates the input ``password`` to exactly 32 bytes. 
    
    - If ``password`` is longer than 32 bytes, it shall be truncated. 
    - If ``password`` is shorter than 32 bytes, it shall be padded by appending data \
    from :const:`.PASSWORD_PADDING` as needed.
    """
    return password[:32] + PASSWORD_PADDING[: 32 - len(password)]


def get_value_from_bytes(contents: PdfHexString | bytes) -> bytes:
    """Returns the inner value of ``contents`` if it is an instance of :class:`.PdfHexString`,
    otherwise returns ``contents``"""
    return contents.value if isinstance(contents, PdfHexString) else contents


class StandardSecurityHandler:
    """An implementation of ``§ 7.6.4 Standard security handler``

    The Standard security handler includes access permissions and allows up to 2 passwords:
    the owner password which has all permissions and the user password which should only
    have the permissions specified by the document.
    """

    def __init__(self, encryption: PdfDictionary, ids: list[PdfHexString | bytes]) -> None:
        """
        Arguments:
            encryption (:class:`.PdfDictionary`):
                The Standard encryption dictionary specified in the document's trailer.
                (``§ 7.6.4 Standard encryption dictionary``)

            ids (:class:`PdfArray[PdfHexString | bytes]`).
                The ID array specified in the document's trailer.
        """
        self.encryption = encryption
        self.ids = ids

    @property
    def key_length(self) -> int:
        """The length of the encryption key in bytes"""
        return self.encryption.get("Length", 40) // 8

    def compute_encryption_key(self, password: bytes) -> bytes:
        """Computes an encryption key from ``password`` according to Algorithm 2 in
        ``§ 7.6.4.3.1 File encryption key algorithm``"""
        # a)
        padded_password = pad_password(password)  # a)

        # b)
        psw_hash = md5(padded_password)
        # c)
        psw_hash.update(get_value_from_bytes(self.encryption["O"]))
        # d)
        psw_hash.update(self.encryption["P"].to_bytes(4, "little", signed=True))
        # e)
        psw_hash.update(get_value_from_bytes(self.ids[0]))

        # f)
        if self.encryption.get("V", 0) >= 4 and not self.encryption.get("EncryptMetadata", True):
            psw_hash.update(b"\xff\xff\xff\xff")

        # g) and h)
        if self.encryption["R"] >= 3:
            for _ in range(50):
                psw_hash = md5(psw_hash.digest()[: self.key_length])

        # i)
        return psw_hash.digest()[: self.key_length]

    def compute_owner_password(self, owner_password: bytes, user_password: bytes) -> bytes:
        """Computes the O (``owner_password``) value in the Encrypt dictionary according
        to Algorithm 3 in ``§ 7.6.4.4 Password algorithms``.

        As a fallback if there is no owner password, ``user_password`` is also specified.
        """
        # a)
        padded = pad_password(owner_password or user_password)
        # b)
        owner_digest = md5(padded).digest()
        # c)
        if self.encryption["R"] >= 3:
            for _ in range(50):
                owner_digest = md5(owner_digest).digest()

        # d)
        owner_cipher = owner_digest[: self.key_length]

        # e)
        padded_user_psw = pad_password(user_password)
        # f)
        arc4 = self._get_provider("ARC4")
        owner_crypt = arc4(owner_cipher).encrypt(padded_user_psw)

        # g)
        if self.encryption["R"] >= 3:
            for i in range(1, 20):
                owner_crypt = arc4(bytearray(b ^ i for b in owner_cipher)).encrypt(owner_crypt)

        # h)
        return owner_crypt

    def compute_user_password(self, password: bytes) -> bytes:
        """Computes the U (user password) value in the Encrypt dictionary according to
        Algorithm 4 (rev. 2) and Algorithm 5 (rev. 3 and 4) in ``§ 7.6.4.4 Password algorithms``.
        """
        # 4 & 5. a)
        encr_key = self.compute_encryption_key(password)

        arc4 = self._get_provider("ARC4")

        if self.encryption["R"] == 2:
            # 4. b) and c)
            padding_crypt = arc4(encr_key).encrypt(PASSWORD_PADDING)
            return padding_crypt
        else:
            # 5. b) and c)
            padded_id_hash = md5(PASSWORD_PADDING + get_value_from_bytes(self.ids[0]))
            # 5. d)
            user_cipher = arc4(encr_key).encrypt(padded_id_hash.digest())

            # 5. e)
            for i in range(1, 20):
                user_cipher = arc4(bytearray(b ^ i for b in encr_key)).encrypt(user_cipher)

            # 5. f)
            return pad_password(user_cipher)

    def authenticate_user_password(self, password: bytes) -> tuple[bytes, bool]:
        """Authenticates the provided user ``password`` according to Algorithms 6
        (based on Algos. 4 and 5) in ``§ 7.6.4.4 Password Algorithms``.

        Returns:
            A tuple of two values: the encryption key that should decrypt the document and
            whether authentication was successful.
        """
        # first step from Algorithms 4 and 5
        encryption_key = self.compute_encryption_key(password)
        stored_password = get_value_from_bytes(self.encryption["U"])

        arc4 = self._get_provider("ARC4")

        if self.encryption["R"] == 2:
            # Algorithm 4: b) and c)
            user_cipher = arc4(encryption_key).encrypt(PASSWORD_PADDING)
            # last step in Algorithm 6
            return (encryption_key, True) if stored_password == user_cipher else (b"", False)
        else:
            # Algorithm 5: b) and c)
            padded_id_hash = md5(PASSWORD_PADDING + get_value_from_bytes(self.ids[0]))
            # Algorithm 5: d)
            user_cipher = arc4(encryption_key).encrypt(padded_id_hash.digest())

            # Algorithm 5: e)
            for i in range(1, 20):
                user_cipher = arc4(bytearray(b ^ i for b in encryption_key)).encrypt(user_cipher)

            # last step in Algorithm 6 for revisions 3 or greater
            return (
                (encryption_key, True) if stored_password[:16] == user_cipher[:16] else (b"", False)
            )

    def authenticate_owner_password(self, password: bytes) -> tuple[bytes, bool]:
        """Authenticates the provided owner ``password`` (or user ``password`` if none)
        according to Algorithm 7 (based on Algo. 3) in ``§ 7.6.4.4 Password Algorithms``.

        Returns:
            A tuple of two values: the encryption key that should decrypt the document and
            whether authentication was successful.
        """
        # (a) to (d) in Algorithm 3
        padded_password = pad_password(password)
        digest = md5(padded_password).digest()
        if self.encryption["R"] >= 3:
            for _ in range(50):
                digest = md5(digest).digest()

        cipher_key = digest[: self.key_length]
        user_cipher = get_value_from_bytes(self.encryption["O"])

        arc4 = self._get_provider("ARC4")
        # Algorithm 7
        if self.encryption["R"] == 2:
            user_cipher = arc4(user_cipher).decrypt(user_cipher)
        else:
            for i in range(19, -1, -1):
                user_cipher = arc4(bytearray(b ^ i for b in cipher_key)).encrypt(user_cipher)

        return self.authenticate_user_password(user_cipher)

    _Encryptable = Union[PdfStream, PdfHexString, bytes]

    def compute_object_crypt(
        self,
        encryption_key: bytes,
        contents: _Encryptable,
        reference: PdfReference,
        *,
        crypt_filter: PdfDictionary | None = None,
    ) -> tuple[CryptMethod, bytes, bytes]:
        """Computes all needed parameters to encrypt or decrypt ``contents`` according to
        ``§ 7.6.3.1 Algorithm 1: Encryption of data using the RC4 and AES algorithms``.

        This algorithm is only applicable for Encrypt versions 1-4 (deprecated in PDF 2.0).
        Encrypt version 5 uses a simpler algorithm described in `§ 7.6.3.2 Algorithm 1.A`.

        Arguments:
            encryption_key (bytes):
                An encryption key generated by :meth:`.compute_encryption_key`

            contents (:class:`.PdfStream` | :class:`.PdfHexString` | bytes):
                The contents to encrypt/decrypt. The type of object will determine what
                crypt filter will be used for decryption (StmF for streams, StrF for
                hex and literal strings).

            reference (:class:`.PdfReference`):
                The reference of either the object itself (in the case of a stream) or
                the object containing it (in the case of a string).

            crypt_filter (:class:`.PdfDictionary`, optional):
                The specific crypt filter to be referenced when decrypting the document.
                If not specified, the default for this type of ``contents`` will be used.

        Returns:
            A tuple of 3 values: the crypt method to apply (AES-CBC or ARC4),
            the key to use with this method, and the data to encrypt/decrypt.
        """
        # NOTE: step a) is satisfied by the "reference" parameter

        # b)
        generation = reference.generation.to_bytes(4, "little")
        object_number = reference.object_number.to_bytes(4, "little")

        extended_key = encryption_key + object_number[:3] + generation[:2]

        method = (
            self._get_cfm_method(crypt_filter) if crypt_filter else self._get_crypt_method(contents)
        )
        if method == "AESV2":
            extended_key += bytes([0x73, 0x41, 0x6C, 0x54])

        # c)
        crypt_key = md5(extended_key).digest()[: self.key_length + 5][:16]

        if isinstance(contents, PdfStream):
            data = contents.raw
        elif isinstance(contents, PdfHexString):
            data = contents.value
        elif isinstance(contents, bytes):
            data = contents
        else:
            raise TypeError("contents arg must be a PDF stream or string")

        return (method, crypt_key, data)

    def encrypt_object(
        self,
        encryption_key: bytes,
        contents: _Encryptable,
        reference: PdfReference,
        *,
        crypt_filter: PdfDictionary | None = None,
    ) -> bytes:
        """Encrypts the specified ``contents`` according to Algorithm 1 in
        ``§ 7.6.3 General Encryption Algorithm``.

        For details on arguments, please see :meth:`.compute_object_crypt`"""

        crypt_method, key, decrypted = self.compute_object_crypt(
            encryption_key, contents, reference, crypt_filter=crypt_filter
        )

        return self._get_provider(crypt_method)(key).encrypt(decrypted)

    def decrypt_object(
        self,
        encryption_key: bytes,
        contents: _Encryptable,
        reference: PdfReference,
        *,
        crypt_filter: PdfDictionary | None = None,
    ) -> bytes:
        """Decrypts the specified ``contents`` according to Algorithm 1 in
        ``§ 7.6.3 General Encryption Algorithm``.

        For details on arguments, please see :meth:`.compute_object_crypt`"""

        crypt_method, key, encrypted = self.compute_object_crypt(
            encryption_key, contents, reference, crypt_filter=crypt_filter
        )

        return self._get_provider(crypt_method)(key).decrypt(encrypted)

    def _get_provider(self, name: str) -> type[CryptProvider]:
        provider = CRYPT_PROVIDERS.get(name)
        if provider is None:
            raise NotImplementedError(
                f"Missing crypt provider for {name}. Register in CRYPT_PROVIDERS or install a compatible module."
            )
        return provider

    CryptMethod = Literal["Identity", "ARC4", "AESV2"]

    def _get_crypt_method(self, contents: _Encryptable) -> CryptMethod:
        if self.encryption.get("V", 0) != 4:
            # ARC4 is assumed given that can only be specified if V = 4. It is definitely
            # not Identity because the document wouldn't be encrypted in that case.
            return "ARC4"

        if isinstance(contents, PdfStream):
            cf_name = self.encryption.get("StmF", PdfName(b"Identity"))
        elif isinstance(contents, (bytes, PdfHexString)):
            cf_name = self.encryption.get("StrF", PdfName(b"Identity"))
        else:
            raise TypeError("contents arg must be a PDF stream or string")

        if cf_name.value == b"Identity":
            return "Identity"  # No processing needed

        crypt_filters = self.encryption.get("CF", {})
        crypter = crypt_filters.get(cf_name.value.decode(), {})

        return self._get_cfm_method(crypter)

    def _get_cfm_method(self, crypt_filter: PdfDictionary) -> CryptMethod:
        cf_name = crypt_filter.get("CFM", PdfName(b"Identity"))
        if cf_name.value == b"Identity":
            return "Identity"
        elif cf_name.value == b"AESV2":
            return "AESV2"
        elif cf_name.value == b"V2":
            return "ARC4"

        raise ValueError(f"Unknown crypt filter for Standard security handler: {cf_name.value!r}")



from __future__ import annotations

from collections import UserDict, UserList
from typing import SupportsIndex, cast, overload

from typing_extensions import TypeVar

from ...exceptions import PdfResolutionError
from .base import PdfObject, PdfReference

DictKey = TypeVar("DictKey", default=str)
DictVal = TypeVar("DictVal", default=PdfObject)


class PdfDictionary(UserDict[DictKey, DictVal]):
    """An associative table containing pairs of objects (entries) where each entry is
    composed of a key (a name object) and a value (any PDF object)
    (``§ 7.3.7 Dictionary objects``).

    :class:`PdfDictionary` is effectively a Python dictionary. Its keys are strings and
    its values are any PDF object. The main difference from a typical dictionary is that
    PdfDictionary automatically resolves references when indexing.

    The underlying data in unresolved form is stored in :attr:`.PdfDictionary.data`.
    """

    def __getitem__(self, key: DictKey) -> DictVal:
        item = self.data[key]
        if isinstance(item, PdfReference):
            try:
                return cast(DictVal, item.get())
            except PdfResolutionError:
                pass

        return cast(DictVal, item)

    def __setitem__(self, key: DictKey, value: DictVal | PdfReference[DictVal]) -> None:
        self.data[key] = cast(DictVal, value)


ArrVal = TypeVar("ArrVal", default=PdfObject)


class PdfArray(UserList[ArrVal]):
    """A heterogeneous collection of sequentially arranged items (``§ 7.3.6 Array objects``).

    :class:`PdfArray` is effectively a Python list. The main difference from a typical list
    is that PdfArray automatically resolves references when indexing.

    The underlying data in unresolved form is stored in :attr:`.PdfArray.data`.
    """

    @overload
    def __getitem__(self, i: SupportsIndex) -> ArrVal: ...

    @overload
    def __getitem__(self, i: slice) -> PdfArray[ArrVal]: ...

    def __getitem__(self, i: SupportsIndex | slice) -> ArrVal | PdfArray[ArrVal]:
        item = self.data[i]
        if isinstance(i, slice):
            return PdfArray(cast(list[ArrVal], item))

        if isinstance(item, PdfReference):
            try:
                return cast(ArrVal, item.get())
            except PdfResolutionError:
                pass

        return cast(ArrVal, item)



from __future__ import annotations

from typing import Generator, cast

from pdfnaut.cos.objects.base import parse_text_string
from pdfnaut.objects.catalog import PageLayout, PageMode

from .cos.objects import (
    PdfArray,
    PdfDictionary,
    PdfHexString,
    PdfName,
    PdfReference,
    PdfStream,
)
from .cos.parser import FreeObject, PdfParser, PermsAcquired
from .objects.page import Page
from .objects.trailer import Info


class PdfDocument(PdfParser):
    """A high-level interface over :class:`~.PdfParser`.

    PDF authors who want to work with a document in a high-level way should use
    this interface over ``PdfParser``."""

    @classmethod
    def from_filename(cls, path: str, *, strict: bool = False) -> PdfDocument:
        """Loads a PDF document from a file ``path``."""
        with open(path, "rb") as fp:
            return PdfDocument(fp.read(), strict=strict)

    def __init__(self, data: bytes, *, strict: bool = False) -> None:
        super().__init__(data, strict=strict)

        self.parse()

        self.access_level = PermsAcquired.OWNER
        """The current access level of the document, specified as a value from the
        :class:`.PermsAcquired` enum.

        - Owner: Full access to the document. If the document is not encrypted, \
        this is the default value.
        - User: Access to the document under restrictions.
        - None: Document is currently encrypted.
        """

        # some files use an empty string as a password
        if self.has_encryption:
            self.access_level = self.decrypt("")

    @property
    def has_encryption(self) -> bool:
        """Whether this document includes encryption."""
        return "Encrypt" in self.trailer

    @property
    def catalog(self) -> PdfDictionary:
        """The root of the document's object hierarchy, including references to pages,
        outlines, destinations, and other core attributes of a PDF document.

        For details on the contents of the catalog, see ``§ 7.7.2 Document Catalog``.
        """
        return cast(PdfDictionary, self.trailer["Root"])

    @catalog.setter
    def catalog(self, value: PdfDictionary) -> None:
        root_ref = cast(PdfReference, self.trailer.data["Root"])
        self.objects[root_ref.object_number] = value

    @property
    def info(self) -> Info | None:
        """The ``Info`` entry in the catalog which includes document-level information
        described in ``§ 14.3.3 Document information dictionary``.

        Some documents may specify a metadata stream rather than an Info entry. This can be
        accessed with :attr:`.PdfDocument.metadata`. PDF 2.0 deprecates all keys of this
        entry except for ``CreationDate`` and ``ModDate``.
        """
        if "Info" not in self.trailer:
            return

        return Info.from_dict(cast(PdfDictionary, self.trailer["Info"]))

    @info.setter
    def info(self, value: Info | None) -> None:
        info_ref = cast("PdfReference | None", self.trailer.data.get("Info"))

        # A new docinfo object will be created
        if info_ref is None and value is not None:
            new_object = max(self.objects) + 1
            self.objects[new_object] = PdfDictionary(**value.data)
            self.trailer.data["Info"] = PdfReference(new_object, 0).with_resolver(self.get_object)
        # A docinfo object will be set
        elif info_ref and isinstance(value, Info):
            self.objects[info_ref.object_number] = PdfDictionary(**value.data)
        # A docinfo object will be removed
        elif info_ref:
            self.objects[info_ref.object_number] = FreeObject()
            self.trailer.data.pop("Info", None)

    @property
    def pdf_version(self) -> str:
        """The version of the PDF standard used in this document.

        The version of a PDF may be identified by either its header or the Version entry
        in the catalog. If the Version entry is absent or the header specifies a later
        version, the header version is returned. Otherwise, the Version entry is returned.
        """
        header_version = self.header_version
        catalog_version = cast("PdfName | None", self.catalog.get("Version"))

        if not catalog_version:
            return header_version

        return max((header_version, catalog_version.value.decode()))

    @property
    def metadata(self) -> PdfStream | None:
        """The Metadata entry of the catalog which includes document-level metadata
        stored as XMP."""
        if "Metadata" not in self.catalog:
            return

        return cast(PdfStream, self.catalog["Metadata"])

    @property
    def page_tree(self) -> PdfDictionary:
        """The document's page tree. See ``§ 7.7.3 Page Tree``.

        For iterating over the pages of a PDF, prefer :attr:`.PdfDocument.flattened_pages`.
        """
        return cast(PdfDictionary, self.catalog["Pages"])

    @property
    def outline_tree(self) -> PdfDictionary | None:
        """The document's outlines commonly referred to as bookmarks.

        See ``§ 12.3.3 Document Outline``."""
        return cast("PdfDictionary | None", self.catalog.get("Outlines"))

    def decrypt(self, password: str) -> PermsAcquired:
        self.access_level = super().decrypt(password)
        return self.access_level

    def _flatten_pages(self, *, parent: PdfDictionary | None = None) -> Generator[Page, None, None]:
        root = cast(PdfDictionary, parent or self.page_tree)

        for page in cast(PdfArray[PdfDictionary], root["Kids"]):
            type_ = cast(PdfName, page["Type"])
            if type_.value == b"Pages":
                yield from self._flatten_pages(parent=page)
            elif type_.value == b"Page":
                yield Page.from_dict(page)

    @property
    def flattened_pages(self) -> Generator[Page, None, None]:
        """A generator suitable for iterating over the pages of a PDF."""
        return self._flatten_pages()

    @property
    def page_layout(self) -> PageLayout:
        """The page layout to use when opening the document. May be one of the following
        values:

        - SinglePage: Display one page at a time (default).
        - OneColumn: Display the pages in one column.
        - TwoColumnLeft: Display the pages in two columns, with odd-numbered pages
          on the left.
        - TwoColumnRight: Display the pages in two columns, with odd-numbered pages
          on the right.
        - TwoPageLeft: Display the pages two at a time, with odd-numbered
          pages on the left (PDF 1.5).
        - TwoPageRight: Display the pages two at a time, with odd-numbered
          pages on the right (PDF 1.5).
        """
        if "PageLayout" not in self.catalog:
            return "SinglePage"

        return cast(PageLayout, cast(PdfName, self.catalog["PageLayout"]).value.decode())

    @property
    def page_mode(self) -> PageMode:
        """Value specifying how the document shall be displayed when opened:

        - UseNone: Neither document outline nor thumbnail images visible (default).
        - UseOutlines: Document outline visible.
        - UseThumbs: Thumbnail images visible.
        - FullScreen: Full-screen mode, with no menu bar, window controls, or any
          other window visible.
        - UseOC: Optional content group panel visible (PDF 1.5).
        - UseAttachments: Attachments panel visible (PDF 1.6).
        """
        if "PageMode" not in self.catalog:
            return "UseNone"

        return cast(PageMode, cast(PdfName, self.catalog["PageMode"]).value.decode())

    @property
    def language(self) -> str | None:
        """A language identifier that shall specify the natural language for all text in
        the document except where overridden by language specifications for structure
        elements or marked content (``§ 14.9.2 Natural language specification``).
        If this entry is absent, the language shall be considered unknown."""
        if "Lang" not in self.catalog:
            return

        return parse_text_string(cast("PdfHexString | bytes", self.catalog["Lang"]))



"""
pdfnaut is a Python library for reading and writing PDFs at a low level.
"""

from __future__ import annotations

from .cos import PdfParser, PdfSerializer, PdfTokenizer
from .document import PdfDocument

__all__ = ("PdfParser", "PdfTokenizer", "PdfSerializer", "PdfDocument")

__name__ = "pdfnaut"
__version__ = "0.7.0"
__description__ = "Explore PDFs with ease"
__license__ = "Apache 2.0"



from __future__ import annotations

import datetime
import re
from typing import NamedTuple

from typing_extensions import Self


class PdfDate(NamedTuple):
    """A date value stored in a PDF document (``§ 7.9.4 Dates``)"""

    year: int
    month: int = 1
    day: int = 1
    hour: int = 0
    minute: int = 0
    second: int = 0
    offset_hour: int = 0
    offset_minute: int = 0

    @classmethod
    def from_pdf(cls, date: str) -> Self:
        """Creates a :class:`PdfDate` from a PDF date string (for example: ``D:20010727133720``)."""
        # pdf 1.7 (and below) dates may end with an apostrophe
        if date.endswith("'"):
            date = date[:-1]

        pattern = re.compile(
            r"""^D:(?P<year>\d{4})(?P<month>\d{2})?(?P<day>\d{2})? # date
                    (?P<hour>\d{2})?(?P<minute>\d{2})?(?P<second>\d{2})? # time
                    (?P<offset>[-+Z])?(?P<offset_hour>\d{2})?(?P<offset_minute>'\d{2})?$ # offset
            """,
            re.X,
        )

        mat = pattern.match(date)
        if not mat:
            raise ValueError(f"Invalid date format: {date!r}")

        offset_sign = mat.group("offset")
        if offset_sign is None or offset_sign == "Z":
            offset_hour = 0
            offset_minute = 0
        else:
            offset_hour = int(mat.group("offset_hour") or 0)
            offset_minute = int((mat.group("offset_minute") or "'0")[1:])

        if offset_sign == "-":
            offset_hour = -offset_hour

        return cls(
            year=int(mat.group("year")),
            month=int(mat.group("month") or 1),
            day=int(mat.group("day") or 1),
            hour=int(mat.group("hour") or 0),
            minute=int(mat.group("minute") or 0),
            second=int(mat.group("second") or 0),
            offset_hour=offset_hour,
            offset_minute=offset_minute,
        )

    @classmethod
    def from_datetime(cls, date: datetime.datetime) -> Self:
        """Creates a :class:`PdfDate` from a :class:`datetime.datetime` object."""
        if date.utcoffset() is None:
            date = date.astimezone()

        if (offset := date.utcoffset()) is not None:
            offset_hours, offset_seconds = divmod(offset.total_seconds(), 3600)
        else:
            offset_hours, offset_seconds = 0, 0

        return cls(
            date.year,
            date.month,
            date.day,
            date.hour,
            date.minute,
            date.second,
            int(offset_hours),
            int(offset_seconds / 60),
        )

    def as_datetime(self) -> datetime.datetime:
        """Returns the :class:`datetime.datetime` equivalent of this date."""
        delta = datetime.timedelta(hours=self.offset_hour, minutes=self.offset_minute)

        return datetime.datetime(
            self.year,
            self.month,
            self.day,
            self.hour,
            self.minute,
            self.second,
            tzinfo=datetime.timezone(delta),
        )

    def as_pdf_string(self) -> str:
        """Returns the PDF equivalent of this date as a string."""
        datestr = f"D:{self.year}"
        datestr += "".join(f"{val:02}" for val in (self.month, self.day) if val != 1)

        datestr += "".join(f"{val:02}" for val in (self.hour, self.minute, self.second) if val != 0)

        if self.offset_hour == 0 and self.offset_minute == 0:
            return datestr + "Z"

        return datestr + "'".join(
            f"{val:02}" for val in (self.offset_hour, self.offset_minute) if val != 0
        )



from __future__ import annotations

import datetime
import enum

from typing_extensions import Self

from .fields import DateField, EnumField, PdfDictionary, TextStringField


class TrappedState(enum.Enum):
    No = 0
    """Document has not been trapped."""
    Yes = 1
    """Document has been already trapped."""
    Unknown = 2
    """Unknown whether document is trapped partly, fully, or at all."""


class Info(PdfDictionary):
    """The document information dictionary (``§ 14.3.3 Document information dictionary``).

    It represents document-level metadata and is stored in the trailer. Since PDF 2.0,
    most of its function has been superseded by a document-level Metadata stream, with
    exception of the CreationDate and ModDate keys.
    """

    title = TextStringField("Title")
    """The document's title."""

    author = TextStringField("Author")
    """The name of the person who created the document."""

    subject = TextStringField("Subject")
    """The subject or topic of the document."""

    keywords = TextStringField("Keywords")
    """Keywords associated with the document."""

    creator = TextStringField("Creator")
    """If the document was converted to PDF from another format (ex. DOCX), the name of 
    the PDF processor that created the original document from which it was converted 
    (ex. Microsoft Word)."""

    producer = TextStringField("Producer")
    """If the document was converted to PDF from another format (ex. PostScript), the name of 
    the PDF processor that converted it to PDF (ex. Adobe Distiller)."""

    creation_date_raw = TextStringField("CreationDate")
    """The date and time the document was created, as a text string."""

    modify_date_raw = TextStringField("ModDate")
    """The date and time the document was most recently modified, as a text string."""

    creation_date = DateField("CreationDate")
    """The date and time the document was created, in human-readable form."""

    modify_date = DateField("ModDate")
    """The date and time the document was most recently modified, in human-readable form."""

    trapped = EnumField(
        "Trapped",
        {"True": TrappedState.Yes, "False": TrappedState.No, "Unknown": TrappedState.Unknown},
        TrappedState.Unknown,
    )
    """A value indicating whether the document has been modified to include trapping 
    information (``§ 14.11.6 Trapping support``)."""

    @classmethod
    def from_dict(cls, mapping: PdfDictionary) -> Self:
        dictionary = cls()
        dictionary.update()
        dictionary.data = mapping.data

        return dictionary

    def __init__(
        self,
        title: str | None = None,
        author: str | None = None,
        subject: str | None = None,
        keywords: str | None = None,
        creator: str | None = None,
        producer: str | None = None,
        creation_date: datetime.datetime | None = None,
        modify_date: datetime.datetime | None = None,
        trapped: TrappedState | None = None,
    ) -> None:
        super().__init__()

        # TODO: I'll rework this to be dataclassy some day.
        self._attrs = {
            "title": title,
            "author": author,
            "subject": subject,
            "keywords": keywords,
            "creator": creator,
            "producer": producer,
            "creation_date": creation_date,
            "modify_date": modify_date,
            "trapped": trapped,
        }

        for name, value in self._attrs.items():
            if value is not None:
                setattr(self, name, value)

    def __repr__(self) -> str:
        attributes = ", ".join(
            f"{attr}={value!r}"
            for attr in self._attrs
            if (value := getattr(self, attr, None)) is not None
        )

        return f"{self.__class__.__name__}({attributes})"
